# Copyright 2016:
# * Jim Unroe KC9HI, <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0010;
struct {
  lbcd rxfreq[4];
  lbcd txfreq[4];
  ul16 rxtone;
  ul16 txtone;
  u8 unknown1:2,
     dtmf:1,          // DTMF
     unknown2:1,
     bcl:1,           // Busy Channel Lockout
     unknown3:3;
  u8 unknown4:1,
     scan:1,          // Scan Add
     highpower:1,     // TX Power Level
     wide:1,          // BandWidth
     unknown5:4;
  u8 unknown6[2];
} memory[128];

#seekto 0x0E17;
struct {
  u8 displayab:1,  // Selected Display
     unknown1:6,
     unknown2:1;
} settings1;

#seekto 0x0E22;
struct {
  u8 squelcha;      // menu 02a Squelch Level              0xe22
  u8 unknown1;
  u8 tdrab;         //          TDR A/B                    0xe24
  u8 roger;         // menu 20  Roger Beep                 0xe25
  u8 timeout;       // menu 16  TOT                        0xe26
  u8 vox;           // menu 05  VOX                        0xe27
  u8 unknown2;
  u8 mdfb;          // menu 27b Memory Display Format B    0xe37
  u8 dw;            // menu 37  DW                         0xe2a
  u8 tdr;           // menu 29  Dual Watch                 0xe2b
  u8 voice;         // menu 03  Voice Prompts              0xe2c
  u8 beep;          // menu 01  Key Beep                   0xe2d
  u8 ani;           // menu 30  ANI                        0xe2e
  u8 unknown3[4];
  u8 pttdly;        // menu 31  PTT-ID Delay               0xe33
  u8 unknown4;
  u8 dtmfst;        // menu 33  DTMF Side Tone             0xe35
  u8 toa;           // menu 15  TOT Pre-Alert              0xe36
  u8 mdfa;          // menu 27a Memory Display Format A    0xe37
  u8 screv;         // menu 09  Scan Resume Method         0xe38
  u8 pttid;         // menu 32  PTT-ID Enable              0xe39
  u8 ponmsg;        // menu 36  Power-on Message           0xe3a
  u8 pf1;           // menu 28  Programmable Function Key  0xe3b
  u8 unknown5;
  u8 wtled;         // menu 17  Standby LED Color          0xe3d
  u8 rxled;         // menu 18  RX LED Color               0xe3e
  u8 txled;         // menu 19  TX LED Color               0xe3f
  u8 unknown6;
  u8 autolk;        // menu 06  Auto Key Lock              0xe41
  u8 squelchb;      // menu 02b Squelch Level              0xe42
  u8 control;       //          Control Code               0xe43
  u8 unknown7;
  u8 ach;           //          Selected A channel Number  0xe45
  u8 unknown8[4];
  u8 password[6];   //          Control Password           0xe4a-0xe4f
  u8 unknown9[7];
  u8 code[3];       //          PTT ID Code                0xe57-0xe59
  u8 vfomr;         //          Frequency/Channel Modevel  0xe5a
  u8 keylk;         //          Key Lock                   0xe5b
  u8 unknown10[2];
  u8 prioritych;    //          Priority Channel           0xe5e
  u8 bch;           //          Selected B channel Number  0xe5f
} settings;

struct vfo {
  u8 unknown0[8];
  u8 freq[8];
  u8 offset[6];
  ul16 rxtone;
  ul16 txtone;
  u8 unused0:7,
     band:1;
  u8 unknown3;
  u8 unknown4:2,
     sftd:2,
     scode:4;
  u8 unknown5;
  u8 unknown6:1,
     step:3,
     unknown7:4;
  u8 txpower:1,
     widenarr:1,
     unknown8:6;
};

#seekto 0x0F10;
struct {
  struct vfo a;
  struct vfo b;
} vfo;

#seekto 0x1010;
struct {
  u8 name[6];
  u8 unknown[10];
} names[128];

"""

# #### MAGICS #########################################################

# TDXone TD-Q8A magic string
MSTRING_TDQ8A = "\x02PYNCRAM"

LIST_DTMF = ["QT", "QT+DTMF"]
LIST_VOICE = ["Off", "Chinese", "English"]
LIST_OFF1TO9 = ["Off"] + list("123456789")
LIST_OFF1TO10 = LIST_OFF1TO9 + ["10"]
LIST_RESUME = ["Time Operated(TO)", "Carrier Operated(CO)", "Search(SE)"]
LIST_COLOR = ["Off", "Blue", "Orange", "Purple"]
LIST_MODE = ["Channel", "Frequency", "Name"]
LIST_PF1 = ["Off", "Scan", "Lamp", "FM Radio", "Alarm"]
LIST_OFF1TO30 = ["OFF"] + ["%s" % x for x in range(1, 31)]
LIST_DTMFST = ["Off", "DTMF Sidetone", "ANI Sidetone", "DTMF+ANI Sidetone"]
LIST_PONMSG = ["Full", "Welcome", "Battery Voltage"]
LIST_TIMEOUT = ["Off"] + ["%s sec" % x for x in range(15, 615, 15)]
LIST_PTTID = ["BOT", "EOT", "Both"]
LIST_ROGER = ["Off"] + LIST_PTTID
LIST_PRIORITY = ["Off"] + ["%s" % x for x in range(1, 129)]
LIST_WORKMODE = ["Frequency", "Channel"]
LIST_AB = ["A", "B"]

LIST_ALMOD = ["Site", "Tone", "Code"]
LIST_BANDWIDTH = ["Wide", "Narrow"]
LIST_DELAYPROCTIME = ["%s ms" % x for x in range(100, 4100, 100)]
LIST_DTMFSPEED = ["%s ms" % x for x in range(50, 2010, 10)]
LIST_OFFAB = ["Off"] + LIST_AB
LIST_RESETTIME = ["%s ms" % x for x in range(100, 16100, 100)]
LIST_SCODE = ["%s" % x for x in range(1, 16)]
LIST_RPSTE = ["Off"] + ["%s" % x for x in range(1, 11)]
LIST_SAVE = ["Off", "1:1", "1:2", "1:3", "1:4"]
LIST_SHIFTD = ["Off", "+", "-"]
LIST_STEDELAY = ["Off"] + ["%s ms" % x for x in range(100, 1100, 100)]
LIST_TXPOWER = ["High", "Low"]
LIST_DTMF_SPECIAL_DIGITS = ["*", "#", "A", "B", "C", "D"]
LIST_DTMF_SPECIAL_VALUES = [0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x00]

CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ?+-*"
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 25.0]
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5),
                chirp_common.PowerLevel("Low", watts=1)]
VALID_BANDS = [(136000000, 174000000),
               (400000000, 520000000)]


def _rawrecv(radio, amount):
    """Raw read from the radio device"""
    data = ""
    try:
        data = radio.pipe.read(amount)
    except:
        msg = "Generic error reading data from radio; check your cable."
        raise errors.RadioError(msg)

    if len(data) != amount:
        msg = "Error reading data from radio: not the amount of data we want."
        raise errors.RadioError(msg)

    return data


def _rawsend(radio, data):
    """Raw send to the radio device"""
    try:
        radio.pipe.write(data)
    except:
        raise errors.RadioError("Error sending data to radio")


def _make_frame(cmd, addr, length, data=""):
    """Pack the info in the header format"""
    frame = struct.pack(">BHB", ord(cmd), addr, length)
    # add the data if set
    if len(data) != 0:
        frame += data
    # return the data
    return frame


def _recv(radio, addr, length):
    """Get data from the radio """
    # read 4 bytes of header
    hdr = _rawrecv(radio, 4)

    # read data
    data = _rawrecv(radio, length)

    # DEBUG
    LOG.info("Response:")
    LOG.debug(util.hexprint(hdr + data))

    c, a, l = struct.unpack(">BHB", hdr)
    if a != addr or l != length or c != ord("W"):
        LOG.error("Invalid answer for block 0x%04x:" % addr)
        LOG.debug("CMD: %s  ADDR: %04x  SIZE: %02x" % (c, a, l))
        raise errors.RadioError("Unknown response from the radio")

    return data


def _do_ident(radio, magic):
    """Put the radio in PROGRAM mode"""
    # set the serial discipline
    radio.pipe.baudrate = 9600

    # send request to enter program mode
    _rawsend(radio, magic)

    ack = _rawrecv(radio, 1)
    if ack != "\x06":
        if ack:
            LOG.debug(repr(ack))
        raise errors.RadioError("Radio did not respond")

    _rawsend(radio, "\x02")

    # Ok, get the response
    ident = _rawrecv(radio, radio._magic_response_length)

    # check if response is OK
    if not ident.startswith("P3107"):
        # bad response
        msg = "Unexpected response, got this:"
        msg += util.hexprint(ident)
        LOG.debug(msg)
        raise errors.RadioError("Unexpected response from radio.")

    # DEBUG
    LOG.info("Valid response, got this:")
    LOG.debug(util.hexprint(ident))

    _rawsend(radio, "\x06")
    ack = _rawrecv(radio, 1)
    if ack != "\x06":
        if ack:
            LOG.debug(repr(ack))
        raise errors.RadioError("Radio refused clone")

    return ident


def _ident_radio(radio):
    for magic in radio._magic:
        error = None
        try:
            data = _do_ident(radio, magic)
            return data
        except errors.RadioError as e:
            print(e)
            error = e
            time.sleep(2)
    if error:
        raise error
    raise errors.RadioError("Radio did not respond")


def _download(radio):
    """Get the memory map"""
    # put radio in program mode
    ident = _ident_radio(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio._mem_size / radio._recv_block_size
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data = ""
    for addr in range(0, radio._mem_size, radio._recv_block_size):
        frame = _make_frame("R", addr, radio._recv_block_size)
        # DEBUG
        LOG.info("Request sent:")
        LOG.debug(util.hexprint(frame))

        # sending the read request
        _rawsend(radio, frame)

        # now we read
        d = _recv(radio, addr, radio._recv_block_size)

        time.sleep(0.05)

        _rawsend(radio, "\x06")

        ack = _rawrecv(radio, 1)
        if ack != "\x06":
            raise errors.RadioError(
                "Radio refused to send block 0x%04x" % addr)

        # aggregate the data
        data += d

        # UI Update
        status.cur = addr / radio._recv_block_size
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    data += radio.MODEL.ljust(8)

    return data


def _upload(radio):
    """Upload procedure"""
    # put radio in program mode
    _ident_radio(radio)

    addr = 0x0f80
    frame = _make_frame("R", addr, radio._recv_block_size)
    # DEBUG
    LOG.info("Request sent:")
    LOG.debug(util.hexprint(frame))

    # sending the read request
    _rawsend(radio, frame)

    # now we read
    d = _recv(radio, addr, radio._recv_block_size)

    time.sleep(0.05)

    _rawsend(radio, "\x06")

    ack = _rawrecv(radio, 1)
    if ack != "\x06":
        raise errors.RadioError(
            "Radio refused to send block 0x%04x" % addr)

    _ranges = radio._ranges

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio._mem_size / radio._send_block_size
    status.msg = "Cloning to radio..."
    radio.status_fn(status)

    # the fun start here
    for start, end in _ranges:
        for addr in range(start, end, radio._send_block_size):
            # sending the data
            data = radio.get_mmap()[addr:addr + radio._send_block_size]

            frame = _make_frame("W", addr, radio._send_block_size, data)

            _rawsend(radio, frame)

            # receiving the response
            ack = _rawrecv(radio, 1)
            if ack != "\x06":
                msg = "Bad ack writing block 0x%04x" % addr
                raise errors.RadioError(msg)

            # UI Update
            status.cur = addr / radio._send_block_size
            status.msg = "Cloning to radio..."
            radio.status_fn(status)


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""

    if len(data) == 0x2008:
        rid = data[0x2000:0x2008]
        return rid.startswith(cls._model)
    else:
        return False


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # if the two freqs are on the same Band this is not a split
            return False

    # if you get here is because the freq pairs are split
    return True


@directory.register
class TDXoneTDQ8A(chirp_common.CloneModeRadio,
                  chirp_common.ExperimentalRadio):
    """TDXone TD-Q8A Radio"""
    VENDOR = "TDXone"
    MODEL = "TD-Q8A"

    _model = b'TD-Q8A'
    _magic = [MSTRING_TDQ8A, MSTRING_TDQ8A, ]
    _magic_response_length = 8
    _fw_ver_start = 0x1EF0
    _recv_block_size = 0x40
    _mem_size = 0x2000

    _ranges = [(0x0010, 0x0810),
               (0x0E20, 0x0E60),
               (0x0F10, 0x0F30),
               (0x1010, 0x1810),
               (0x1F10, 0x1F30)]
    _send_block_size = 0x10

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('The TDXone TD-Q8A driver is a beta version.\n'
             '\n'
             'Please save an unedited copy of your first successful\n'
             'download to a CHIRP Radio Images(*.img) file.'
             )
        rp.pre_download = _(
            "Follow these instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        """Get the radio's features"""

        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = False
        rf.has_rx_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_characters = CHARSET
        rf.valid_name_length = 6
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "->Tone"]
        rf.valid_skips = ["", "S"]
        rf.memory_bounds = (1, 128)
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tuning_steps = STEPS
        rf.valid_bands = VALID_BANDS

        return rf

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = _download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = memmap.MemoryMap(data)
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            _upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def _is_txinh(self, _mem):
        raw_tx = ""
        for i in range(0, 4):
            raw_tx += _mem.txfreq[i].get_raw(asbytes=False)
        return raw_tx == "\xFF\xFF\xFF\xFF"

    def _get_mem(self, number):
        return self._memobj.memory[number - 1]

    def _get_nam(self, number):
        return self._memobj.names[number - 1]

    def get_memory(self, number):
        _mem = self._get_mem(number)
        _nam = self._get_nam(number)

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xff":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rxfreq) * 10

        if self._is_txinh(_mem):
            # TX freq not set
            mem.duplex = "off"
            mem.offset = 0
        else:
            # TX freq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset != 0:
                if _split(self.get_features(), mem.freq, int(
                          _mem.txfreq) * 10):
                    mem.duplex = "split"
                    mem.offset = int(_mem.txfreq) * 10
                elif offset < 0:
                    mem.offset = abs(offset)
                    mem.duplex = "-"
                elif offset > 0:
                    mem.offset = offset
                    mem.duplex = "+"
            else:
                mem.offset = 0

        if _nam.name:
            for char in _nam.name:
                try:
                    mem.name += CHARSET[char]
                except IndexError:
                    break
            mem.name = mem.name.rstrip()

        if _mem.txtone in [0, 0xFFFF]:
            txmode = ""
        elif _mem.txtone >= 0x0258:
            txmode = "Tone"
            mem.rtone = int(_mem.txtone) / 10.0
        else:
            LOG.warn("Bug: txtone is %04x" % _mem.txtone)

        if _mem.rxtone in [0, 0xFFFF]:
            rxmode = ""
        elif _mem.rxtone >= 0x0258:
            rxmode = "Tone"
            mem.ctone = int(_mem.rxtone) / 10.0
        else:
            LOG.warn("Bug: rxtone is %04x" % _mem.rxtone)

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        if not _mem.scan:
            mem.skip = "S"

        mem.power = POWER_LEVELS[1 - _mem.highpower]

        mem.mode = _mem.wide and "FM" or "NFM"

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("dtmf", "DTMF",
                          RadioSettingValueList(LIST_DTMF,
                                                LIST_DTMF[_mem.dtmf]))
        mem.extra.append(rs)

        rs = RadioSetting("bcl", "BCL",
                          RadioSettingValueBoolean(_mem.bcl))
        mem.extra.append(rs)

        return mem

    def _set_mem(self, number):
        return self._memobj.memory[number - 1]

    def _set_nam(self, number):
        return self._memobj.names[number - 1]

    def set_memory(self, mem):
        _mem = self._get_mem(mem.number)
        _nam = self._get_nam(mem.number)

        if mem.empty:
            _mem.set_raw("\xff" * 12 + "\xbf" + "\xff" * 3)
            _nam.set_raw("\xff" * 16)
            return

        _mem.set_raw("\xff" * 12 + "\x9f" + "\xff" * 3)

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw("\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        if _nam.name:
            for i in range(0, 6):
                try:
                    _nam.name[i] = CHARSET.index(mem.name[i])
                except IndexError:
                    _nam.name[i] = 0xFF

        rxmode = txmode = ""
        if mem.tmode == "Tone":
            _mem.txtone = int(mem.rtone * 10)
            _mem.rxtone = 0
        elif mem.tmode == "TSQL":
            _mem.txtone = int(mem.ctone * 10)
            _mem.rxtone = int(mem.ctone * 10)
        elif mem.tmode == "Cross":
            txmode, rxmode = mem.cross_mode.split("->", 1)
            if txmode == "Tone":
                _mem.txtone = int(mem.rtone * 10)
            else:
                _mem.txtone = 0
            if rxmode == "Tone":
                _mem.rxtone = int(mem.ctone * 10)
            else:
                _mem.rxtone = 0
        else:
            _mem.rxtone = 0
            _mem.txtone = 0

        _mem.scan = mem.skip != "S"
        _mem.wide = mem.mode == "FM"

        _mem.highpower = mem.power == POWER_LEVELS[0]

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic Settings")
        advanced = RadioSettingGroup("advanced", "Advanced Settings")
        top = RadioSettings(basic, advanced, )

        # Basic settings
        rs = RadioSetting("settings.beep", "Beep",
                          RadioSettingValueBoolean(
                              _mem.settings.beep))
        basic.append(rs)

        if _mem.settings.squelcha > 0x09:
            val = 0x00
        else:
            val = _mem.settings.squelcha
        rs = RadioSetting("squelcha", "Squelch Level A",
                          RadioSettingValueInteger(
                              0, 9, _mem.settings.squelcha))
        basic.append(rs)

        if _mem.settings.squelchb > 0x09:
            val = 0x00
        else:
            val = _mem.settings.squelchb
        rs = RadioSetting("squelchb", "Squelch Level B",
                          RadioSettingValueInteger(
                              0, 9, _mem.settings.squelchb))
        basic.append(rs)

        if _mem.settings.voice > 0x02:
            val = 0x01
        else:
            val = _mem.settings.voice
        rs = RadioSetting("settings.voice", "Voice Prompt",
                          RadioSettingValueList(
                              LIST_VOICE, LIST_VOICE[val]))
        basic.append(rs)

        if _mem.settings.vox > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.vox
        rs = RadioSetting("settings.vox", "VOX",
                          RadioSettingValueList(
                              LIST_OFF1TO10, LIST_OFF1TO10[val]))
        basic.append(rs)

        rs = RadioSetting("settings.autolk", "Automatic Key Lock",
                          RadioSettingValueBoolean(_mem.settings.autolk))
        basic.append(rs)

        if _mem.settings.screv > 0x02:
            val = 0x01
        else:
            val = _mem.settings.screv
        rs = RadioSetting("settings.screv", "Scan Resume",
                          RadioSettingValueList(
                              LIST_RESUME, LIST_RESUME[val]))
        basic.append(rs)

        if _mem.settings.toa > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.toa
        rs = RadioSetting("settings.toa", "Time-out Pre-Alert",
                          RadioSettingValueList(
                              LIST_OFF1TO10, LIST_OFF1TO10[val]))
        basic.append(rs)

        if _mem.settings.timeout > 0x28:
            val = 0x03
        else:
            val = _mem.settings.timeout
        rs = RadioSetting("settings.timeout", "Timeout Timer",
                          RadioSettingValueList(
                              LIST_TIMEOUT, LIST_TIMEOUT[val]))
        basic.append(rs)

        rs = RadioSetting("settings.wtled", "Standby LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.wtled]))
        basic.append(rs)

        rs = RadioSetting("settings.rxled", "RX LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.rxled]))
        basic.append(rs)

        rs = RadioSetting("settings.txled", "TX LED Color",
                          RadioSettingValueList(
                              LIST_COLOR, LIST_COLOR[_mem.settings.txled]))
        basic.append(rs)

        rs = RadioSetting("settings.roger", "Roger Beep",
                          RadioSettingValueList(LIST_ROGER, LIST_ROGER[
                              _mem.settings.roger]))
        basic.append(rs)

        rs = RadioSetting("settings.mdfa", "Display Mode (A)",
                          RadioSettingValueList(LIST_MODE, LIST_MODE[
                              _mem.settings.mdfa]))
        basic.append(rs)

        rs = RadioSetting("settings.mdfb", "Display Mode (B)",
                          RadioSettingValueList(LIST_MODE, LIST_MODE[
                              _mem.settings.mdfb]))
        basic.append(rs)

        rs = RadioSetting("settings.pf1", "PF1 Key Assignment",
                          RadioSettingValueList(LIST_PF1, LIST_PF1[
                              _mem.settings.pf1]))
        basic.append(rs)

        rs = RadioSetting("settings.tdr", "Dual Watch(TDR)",
                          RadioSettingValueBoolean(_mem.settings.tdr))
        basic.append(rs)

        rs = RadioSetting("settings.ani", "ANI",
                          RadioSettingValueBoolean(_mem.settings.ani))
        basic.append(rs)

        if _mem.settings.pttdly > 0x0A:
            val = 0x00
        else:
            val = _mem.settings.pttdly
        rs = RadioSetting("settings.pttdly", "PTT ID Delay",
                          RadioSettingValueList(
                              LIST_OFF1TO30, LIST_OFF1TO30[val]))
        basic.append(rs)

        rs = RadioSetting("settings.pttid", "When to send PTT ID",
                          RadioSettingValueList(LIST_PTTID,
                                                LIST_PTTID[
                                                    _mem.settings.pttid]))
        basic.append(rs)

        rs = RadioSetting("settings.dtmfst", "DTMF Sidetone",
                          RadioSettingValueList(LIST_DTMFST, LIST_DTMFST[
                              _mem.settings.dtmfst]))
        basic.append(rs)

        rs = RadioSetting("settings.ponmsg", "Power-On Message",
                          RadioSettingValueList(LIST_PONMSG, LIST_PONMSG[
                              _mem.settings.ponmsg]))
        basic.append(rs)

        rs = RadioSetting("settings.dw", "DW",
                          RadioSettingValueBoolean(_mem.settings.dw))
        basic.append(rs)

        # Advanced settings
        rs = RadioSetting("settings.prioritych", "Priority Channel",
                          RadioSettingValueList(LIST_PRIORITY, LIST_PRIORITY[
                              _mem.settings.prioritych]))
        advanced.append(rs)

        rs = RadioSetting("settings.vfomr", "Work Mode",
                          RadioSettingValueList(LIST_WORKMODE, LIST_WORKMODE[
                              _mem.settings.vfomr]))
        advanced.append(rs)

        dtmfchars = "0123456789"
        _codeobj = _mem.settings.code
        _code = "".join([dtmfchars[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 3, _code, False)
        val.set_charset(dtmfchars)
        rs = RadioSetting("settings.code", "PTT-ID Code", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 3):
                try:
                    code.append(dtmfchars.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.code = code
        rs.set_apply_callback(apply_code, _mem.settings)
        advanced.append(rs)

        _codeobj = _mem.settings.password
        _code = "".join([dtmfchars[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 6, _code, False)
        val.set_charset(dtmfchars)
        rs = RadioSetting("settings.password", "Control Password", val)

        def apply_code(setting, obj):
            code = []
            for j in range(0, 6):
                try:
                    code.append(dtmfchars.index(str(setting.value)[j]))
                except IndexError:
                    code.append(0xFF)
            obj.password = code
        rs.set_apply_callback(apply_code, _mem.settings)
        advanced.append(rs)

        if _mem.settings.tdrab > 0x01:
            val = 0x00
        else:
            val = _mem.settings.tdrab
        rs = RadioSetting("settings.tdrab", "Dual Watch TX Priority",
                          RadioSettingValueList(
                              LIST_AB, LIST_AB[val]))
        advanced.append(rs)

        rs = RadioSetting("settings.keylk", "Key Lock",
                          RadioSettingValueBoolean(_mem.settings.keylk))
        advanced.append(rs)

        rs = RadioSetting("settings.control", "Control Code",
                          RadioSettingValueBoolean(_mem.settings.control))
        advanced.append(rs)

        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                if element.get_name() == "fm_preset":
                    self._set_fm_preset(element)
                else:
                    self.set_settings(element)
                    continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    def _set_fm_preset(self, settings):
        for element in settings:
            try:
                val = element.value
                if self._memobj.fm_presets <= 108.0 * 10 - 650:
                    value = int(val.get_value() * 10 - 650)
                else:
                    value = int(val.get_value() * 10)
                LOG.debug("Setting fm_presets = %s" % (value))
                self._memobj.fm_presets = value
            except Exception:
                LOG.debug(element.get_name())
                raise

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == 0x2008:
            match_size = True

        # testing the model fingerprint
        match_model = model_match(cls, filedata)

        # if match_size and match_model:
        if match_size and match_model:
            return True
        else:
            return False



# YPL parser 1.5

# written by VB.

import re
import sys
import collections

import six


class keyword(str):
    pass


class code(str):
    pass


class ignore(object):
    def __init__(self, regex_text, *args):
        self.regex = re.compile(regex_text, *args)


class _and(object):
    def __init__(self, something):
        self.obj = something


class _not(_and):
    pass


class Name(str):
    def __init__(self, *args):
        self.line = 0
        self.file = ""


class Symbol(list):
    def __init__(self, name, what):
        self.__name__ = name
        self.append(name)
        self.what = what
        self.append(what)

    def __call__(self):
        return self.what

    def __unicode__(self):
        return 'Symbol(' + repr(self.__name__) + ', ' + repr(self.what) + ')'

    def __repr__(self):
        return str(self)


word_regex = re.compile(r"\w+")
rest_regex = re.compile(r".*")

print_trace = False


def u(text):
    return six.text_type(text)


def skip(skipper, text, skipWS, skipComments):
    if skipWS:
        t = text.lstrip()
    else:
        t = text
    if skipComments:
        try:
            while True:
                skip, t = skipper.parseLine(t, skipComments, [], skipWS, None)
                if skipWS:
                    t = t.lstrip()
        except Exception:
            pass
    return t


class parser(object):
    def __init__(self, another=False, p=False):
        self.restlen = -1
        if not (another):
            self.skipper = parser(True, p)
            self.skipper.packrat = p
        else:
            self.skipper = self
        self.lines = None
        self.textlen = 0
        self.memory = {}
        self.packrat = p

    # parseLine():
    #   textline:       text to parse
    #   pattern:        pyPEG language description
    #   resultSoFar:    parsing result so far (default: blank list [])
    #   skipWS:         Flag if whitespace should be skipped (default: True)
    #   skipComments:   Python functions returning pyPEG for matching comments
    #
    #   returns:        pyAST, textrest
    #
    #   raises:         SyntaxError(reason) if textline is detected not
    #                   being in language described by pattern
    #
    #                   SyntaxError(reason) if pattern is an illegal
    #                   language description

    def parseLine(self, textline, pattern, resultSoFar=[],
                  skipWS=True, skipComments=None):
        name = None
        _textline = textline
        _pattern = pattern

        def R(result, text):
            if __debug__:
                if print_trace:
                    try:
                        if _pattern.__name__ != "comment":
                            sys.stderr.write("match: " +
                                             _pattern.__name__ + "\n")
                    except Exception:
                        pass

            if self.restlen == -1:
                self.restlen = len(text)
            else:
                self.restlen = min(self.restlen, len(text))
            res = resultSoFar
            if name and result:
                name.line = self.lineNo()
                res.append(Symbol(name, result))
            elif name:
                name.line = self.lineNo()
                res.append(Symbol(name, []))
            elif result:
                if isinstance(result, list):
                    res.extend(result)
                else:
                    res.extend([result])
            if self.packrat:
                self.memory[(len(_textline), id(_pattern))] = (res, text)
            return res, text

        def syntaxError():
            if self.packrat:
                self.memory[(len(_textline), id(_pattern))] = False
            raise SyntaxError()

        if self.packrat:
            try:
                result = self.memory[(len(textline), id(pattern))]
                if result:
                    return result
                else:
                    raise SyntaxError()
            except Exception:
                pass

        if isinstance(pattern, collections.abc.Callable):
            if __debug__:
                if print_trace:
                    try:
                        if pattern.__name__ != "comment":
                            sys.stderr.write("testing with " +
                                             pattern.__name__ + ": " +
                                             textline[:40] + "\n")
                    except Exception:
                        pass

            if pattern.__name__[0] != "_":
                name = Name(pattern.__name__)

            pattern = pattern()
            if isinstance(pattern, collections.abc.Callable):
                pattern = (pattern,)

        text = skip(self.skipper, textline, skipWS, skipComments)

        pattern_type = type(pattern)

        if pattern_type is str or pattern_type is str:
            if text[:len(pattern)] == pattern:
                text = skip(self.skipper, text[len(pattern):],
                            skipWS, skipComments)
                return R(None, text)
            else:
                syntaxError()

        elif pattern_type is keyword:
            m = word_regex.match(text)
            if m:
                if m.group(0) == pattern:
                    text = skip(self.skipper, text[len(pattern):],
                                skipWS, skipComments)
                    return R(None, text)
                else:
                    syntaxError()
            else:
                syntaxError()

        elif pattern_type is _not:
            try:
                r, t = self.parseLine(text, pattern.obj, [],
                                      skipWS, skipComments)
            except Exception:
                return resultSoFar, textline
            syntaxError()

        elif pattern_type is _and:
            r, t = self.parseLine(text, pattern.obj, [], skipWS, skipComments)
            return resultSoFar, textline

        elif pattern_type is type(word_regex) or pattern_type is ignore:
            if pattern_type is ignore:
                pattern = pattern.regex
            m = pattern.match(text)
            if m:
                text = skip(self.skipper, text[len(m.group(0)):],
                            skipWS, skipComments)
                if pattern_type is ignore:
                    return R(None, text)
                else:
                    return R(m.group(0), text)
            else:
                syntaxError()

        elif pattern_type is tuple:
            result = []
            n = 1
            for p in pattern:
                if isinstance(p, int):
                    n = p
                else:
                    if n > 0:
                        for i in range(n):
                            result, text = self.parseLine(
                                text, p, result, skipWS, skipComments)
                    elif n == 0:
                        if text == "":
                            pass
                        else:
                            try:
                                newResult, newText = self.parseLine(
                                    text, p, result, skipWS, skipComments)
                                result, text = newResult, newText
                            except SyntaxError:
                                pass
                    elif n < 0:
                        found = False
                        while True:
                            try:
                                newResult, newText = self.parseLine(
                                    text, p, result, skipWS, skipComments)
                                result, text, found = newResult, newText, True
                            except SyntaxError:
                                break
                        if n == -2 and not (found):
                            syntaxError()
                    n = 1
            return R(result, text)

        elif pattern_type is list:
            result = []
            found = False
            for p in pattern:
                try:
                    result, text = self.parseLine(text, p, result,
                                                  skipWS, skipComments)
                    found = True
                except SyntaxError:
                    pass
                if found:
                    break
            if found:
                return R(result, text)
            else:
                syntaxError()

        else:
            raise SyntaxError("illegal type in grammar: " + u(pattern_type))

    def lineNo(self):
        if not (self.lines):
            return ""
        if self.restlen == -1:
            return ""
        parsed = self.textlen - self.restlen

        left, right = 0, len(self.lines)

        while True:
            mid = int((right + left) / 2)
            if self.lines[mid][0] <= parsed:
                try:
                    if self.lines[mid + 1][0] >= parsed:
                        try:
                            return u(self.lines[mid + 1][1]) + \
                                   ":" + u(self.lines[mid + 1][2])
                        except Exception:
                            return ""
                    else:
                        left = mid + 1
                except Exception:
                    try:
                        return u(self.lines[mid + 1][1]) + \
                               ":" + u(self.lines[mid + 1][2])
                    except Exception:
                        return ""
            else:
                right = mid - 1
            if left > right:
                return ""


# plain module APIs


def parseLine(textline, pattern, resultSoFar=[], skipWS=True,
              skipComments=None, packrat=False):
    p = parser(p=packrat)
    text = skip(p.skipper, textline, skipWS, skipComments)
    ast, text = p.parseLine(text, pattern, resultSoFar, skipWS, skipComments)
    return ast, text

# parse():
#   language:       pyPEG language description
#   lineSource:     a fileinput.FileInput object
#   skipWS:         Flag if whitespace should be skipped (default: True)
#   skipComments:   Python function which returns pyPEG for matching comments
#   packrat:        use memoization
#   lineCount:      add line number information to AST
#
#   returns:        pyAST
#
#   raises:         SyntaxError(reason), if a parsed line is not in language
#                   SyntaxError(reason), if the language description is illegal


def parse(language, lineSource, skipWS=True, skipComments=None,
          packrat=False, lineCount=True):
    lines, lineNo = [], 0

    while isinstance(language, collections.abc.Callable):
        language = language()

    orig, ld = "", 0
    for line in lineSource:
        if lineSource.isfirstline():
            ld = 1
        else:
            ld += 1
        lines.append((len(orig), lineSource.filename(),
                     lineSource.lineno() - 1))
        orig += u(line)

    textlen = len(orig)

    try:
        p = parser(p=packrat)
        p.textlen = len(orig)
        if lineCount:
            p.lines = lines
        else:
            p.line = None
        text = skip(p.skipper, orig, skipWS, skipComments)
        result, text = p.parseLine(text, language, [], skipWS, skipComments)
        if text:
            raise SyntaxError()

    except SyntaxError:
        parsed = textlen - p.restlen
        textlen = 0
        nn, lineNo, file = 0, 0, ""
        for n, ld, l in lines:
            if n >= parsed:
                break
            else:
                lineNo = l
                nn += 1
                file = ld

        lineNo += 1
        nn -= 1
        lineCont = orig.splitlines()[nn]
        raise SyntaxError("syntax error in " + u(file) + ":" +
                          u(lineNo) + ": " + lineCont)

    return result



#
# Copyright 2012 Filippi Marco <iz3gme.marco@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""FT857 - FT857/US management module"""

from chirp.drivers import ft817
from chirp import chirp_common, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings
import logging
from chirp.util import safe_charset_string

LOG = logging.getLogger(__name__)


@directory.register
class FT857Radio(ft817.FT817Radio):
    """Yaesu FT-857/897"""
    MODEL = "FT-857/897"
    _model = ""

    TMODES = {
        0x04: "Tone",
        0x05: "TSQL",
        # 0x08 : "DTCS Enc", not supported in UI yet
        0x0a: "DTCS",
        0xff: "Cross",
        0x00: "",
    }
    TMODES_REV = dict(list(zip(list(TMODES.values()), list(TMODES.keys()))))

    CROSS_MODES = {
        0x01: "->Tone",
        0x02: "->DTCS",
        0x04: "Tone->",
        0x05: "Tone->Tone",
        0x06: "Tone->DTCS",
        0x08: "DTCS->",
        0x09: "DTCS->Tone",
        0x0a: "DTCS->DTCS",
    }
    CROSS_MODES_REV = dict(list(zip(list(CROSS_MODES.values()),
                                    list(CROSS_MODES.keys()))))

    _memsize = 7341
    # block 9 (140 Bytes long) is to be repeated 40 times
    # should be 42 times but this way I can use original 817 functions
    _block_lengths = [2, 82, 252, 196, 252, 196, 212, 55, 140, 140, 140,
                      38, 176]
    # warning ranges has to be in this exact order
    VALID_BANDS = [(100000, 33000000), (33000000, 56000000),
                   (76000000, 108000000), (108000000, 137000000),
                   (137000000, 164000000), (420000000, 470000000)]

    CHARSET = list(chirp_common.CHARSET_ASCII)
    for i in "\\{|}":
        CHARSET.remove(i)

    MEM_FORMAT = """
        struct mem_struct{
        u8   tag_on_off:1,
            tag_default:1,
            unknown1:3,
            mode:3;
        u8   duplex:2,
            is_duplex:1,
            is_cwdig_narrow:1,
            is_fm_narrow:1,
            freq_range:3;
        u8   skip:1,
            unknokwn1_1:1,
            ipo:1,
            att:1,
            unknown2:4;
        u8   ssb_step:2,
            am_step:3,
            fm_step:3;
        u8   unknown3:3,
            is_split_tone:1,
            tmode:4;
        u8   unknown4:2,
            tx_mode:3,
            tx_freq_range:3;
        u8   unknown5:1,
            unknown_toneflag:1,
            tone:6;
        u8   unknown6:1,
            unknown_rxtoneflag:1,
            rxtone:6;
        u8   unknown7:1,
            dcs:7;
        u8   unknown8:1,
            rxdcs:7;
        ul16 rit;
        u32 freq;
        u32 offset;
        u8   name[8];
        };

        #seekto 0x00;
        struct {
            u16 radioconfig;
            u8  mem_vfo:2,
                m_tune:1,
                home:1,
                pms_tune:1,
                qmb:1,
                mt_qmb:1,
                vfo_ab:1;
            u8  unknown;
            u8  fst:1,
                lock:1,
                nb:1,
                unknown1:2,
                disp:1,
                agc:2;
            u8  vox:1,
                unknown2:1,
                bk:1,
                kyr:1,
                cw_speed_unit:1,
                cw_key_rev:1,
                pwr_meter_mode:2;
            u8  vfo_b_freq_range:4,
                vfo_a_freq_range:4;
            u8  unknown3;
            u8  disp_mode:2,
                unknown4:2,
                disp_contrast:4;
            u8  unknown5:4,
                clar_dial_sel:2,
                beep_tone:2;
            u8  arts_beep:2,
                dial_step:1,
                arts_id:1,
                unknown6:1,
                pkt_rate:1,
                unknown7:2;
            u8  unknown8:2,
                lock_mode:2,
                unknown9:1,
                cw_pitch:3;
            u8  sql_rf_gain:1,
                ars_144:1,
                ars_430:1,
                cw_weight:5;
            u8  cw_delay;
            u8  cw_delay_hi:1,
                cw_sidetone:7;
            u8  unknown10:2,
                cw_speed:6;
            u8  disable_amfm_dial:1,
                vox_gain:7;
            u8  cat_rate:2,
                emergency:1,
                vox_delay:5;
            u8  dig_mode:3,
                mem_group:1,
                unknown11:1,
                apo_time:3;
            u8  dcs_inv:2,
                unknown12:1,
                tot_time:5;
            u8  mic_scan:1,
                ssb_mic:7;
            u8  cw_paddle:1,
                am_mic:7;
            u8  unknown13:1,
                fm_mic:7;
            u8  unknown14:1,
                dig_mic:7;
            u8  extended_menu:1,
                pkt1200:7;
            u8  unknown15:1,
                pkt9600:7;
            i16 dig_shift;
            i16 dig_disp;
            i8  r_lsb_car;
            i8  r_usb_car;
            i8  t_lsb_car;
            i8  t_usb_car;
            u8  unknown16:1,
                menu_item:7;
            u8  unknown17[5];
            u8  unknown18:1,
                mtr_peak_hold:1,
                mic_sel:2,
                cat_lin_tun:2,
                unknown19:1,
                split_tone:1;
            u8  unknown20:1,
                beep_vol:7;
            u8  unknown21:1,
                dig_vox:7;
            u8  ext_menu:1,
                home_vfo:1,
                scan_mode:2,
                scan_resume:4;
            u8  cw_auto_mode:1,
                cw_training:2,
                cw_qsk:3,
                cw_bfo:2;
            u8  dsp_nr:4,
                dsp_bpf:2,
                dsp_mic_eq:2;
            u8  unknown22:3,
                dsp_lpf:5;
            u8  mtr_atx_sel:3,
                unknown23:1,
                dsp_hpf:4;
            u8  unknown24:2,
                disp_intensity:2,
                unknown25:1,
                disp_color:3;
            u8  unknown26:1,
                disp_color_vfo:1,
                disp_color_mtr:1,
                disp_color_mode:1,
                disp_color_memgrp:1,
                unknown27:1,
                disp_color_band:1,
                disp_color_arts:1;
            u8  unknown28:3,
                disp_color_fix:5;
            u8  unknown29:1,
                nb_level:7;
            u8  unknown30:1,
                proc_level:7;
            u8  unknown31:1,
                rf_power_hf:7;
            u8  unknown32:2,
                tuner_atas:3,
                mem_vfo_dial_mode:3;
            u8  pg_a;
            u8  pg_b;
            u8  pg_c;
            u8  pg_acc;
            u8  pg_p1;
            u8  pg_p2;
            u8  unknown33:3,
                xvtr_sel:2,
                unknown33_1:2,
                op_filter1:1;
            u8  unknown34:6,
                tx_if_filter:2;
            u8  unknown35:3,
                xvtr_a_negative:1,
                xvtr_b_negative:1,
                mtr_arx_sel:3;
            u8  beacon_time;
            u8  unknown36[2];
            u8  dig_vox_enable:1,
                unknown37:2,
                scope_peakhold:1,
                scope_width:2,
                proc:1,
                unknown38:1;
            u8  unknown39:1,
                rf_power_6m:7;
            u8  unknown40:1,
                rf_power_vhf:7;
            u8  unknown41:1,
                rf_power_uhf:7;
        } settings;

        #seekto 0x54;
        struct mem_struct vfoa[16];
        struct mem_struct vfob[16];
        struct mem_struct home[4];
        struct mem_struct qmb;
        struct mem_struct mtqmb;
        struct mem_struct mtune;

        #seekto 0x4a9;
        u8 visible[25];
        ul16 pmsvisible;

        #seekto 0x4c4;
        u8 filled[25];
        ul16 pmsfilled;

        #seekto 0x4df;
        struct mem_struct memory[200];
        struct mem_struct pms[10];

        #seekto 0x1bf3;
        u8 arts_idw[10];
        u8 beacon_text1[40];
        u8 beacon_text2[40];
        u8 beacon_text3[40];
        u32 xvtr_a_offset;
        u32 xvtr_b_offset;
        u8 op_filter1_name[4];
        u8 op_filter2_name[4];

        #seekto 0x1CAD;
        struct mem_struct sixtymeterchannels[5];

    """

    _CALLSIGN_CHARSET = [chr(x) for x in list(range(ord("0"), ord("9") + 1)) +
                         list(range(ord("A"), ord("Z") + 1))] + [" ", "/"]
    _CALLSIGN_CHARSET_REV = dict(list(zip(_CALLSIGN_CHARSET,
                                 list(range(0, len(_CALLSIGN_CHARSET))))))
    _BEACON_CHARSET = _CALLSIGN_CHARSET + ["+", "."]
    _BEACON_CHARSET_REV = dict(list(zip(_BEACON_CHARSET,
                               list(range(0, len(_BEACON_CHARSET))))))

    # WARNING Index are hard wired in memory management code !!!
    SPECIAL_MEMORIES = {
        "VFOa-1.8M": -37,
        "VFOa-3.5M": -36,
        "VFOa-5M": -35,
        "VFOa-7M": -34,
        "VFOa-10M": -33,
        "VFOa-14M": -32,
        "VFOa-18M": -31,
        "VFOa-21M": -30,
        "VFOa-24M": -29,
        "VFOa-28M": -28,
        "VFOa-50M": -27,
        "VFOa-FM": -26,
        "VFOa-AIR": -25,
        "VFOa-144": -24,
        "VFOa-430": -23,
        "VFOa-HF": -22,
        "VFOb-1.8M": -21,
        "VFOb-3.5M": -20,
        "VFOb-5M": -19,
        "VFOb-7M": -18,
        "VFOb-10M": -17,
        "VFOb-14M": -16,
        "VFOb-18M": -15,
        "VFOb-21M": -14,
        "VFOb-24M": -13,
        "VFOb-28M": -12,
        "VFOb-50M": -11,
        "VFOb-FM": -10,
        "VFOb-AIR": -9,
        "VFOb-144M": -8,
        "VFOb-430M": -7,
        "VFOb-HF": -6,
        "HOME HF": -5,
        "HOME 50M": -4,
        "HOME 144M": -3,
        "HOME 430M": -2,
        "QMB": -1,
    }
    FIRST_VFOB_INDEX = -6
    LAST_VFOB_INDEX = -21
    FIRST_VFOA_INDEX = -22
    LAST_VFOA_INDEX = -37

    SPECIAL_PMS = {
        "PMS-1L": -47,
        "PMS-1U": -46,
        "PMS-2L": -45,
        "PMS-2U": -44,
        "PMS-3L": -43,
        "PMS-3U": -42,
        "PMS-4L": -41,
        "PMS-4U": -40,
        "PMS-5L": -39,
        "PMS-5U": -38,
    }
    LAST_PMS_INDEX = -47

    SPECIAL_MEMORIES.update(SPECIAL_PMS)

    SPECIAL_MEMORIES_REV = dict(list(zip(list(SPECIAL_MEMORIES.values()),
                                         list(SPECIAL_MEMORIES.keys()))))

    FILTERS = ["CFIL", "FIL1", "FIL2"]
    PROGRAMMABLEOPTIONS = [
        "MFa:A/B",              "MFa:A=B",          "MFa:SPL",
        "MFb:MW",               "MFb:SKIP/MCLR",    "MFb:TAG",
        "MFc:STO",              "MFc:RCL",          "MFc:PROC",
        "MFd:RPT",              "MFd:REV",          "MFd:VOX",
        "MFe:TON/ENC",          "MFe:TON/DEC",      "MFe:TDCH",
        "MFf:ARTS",             "MFf:SRCH",         "MFf:PMS",
        "MFg:SCN",              "MFg:PRI",          "MFg:DW",
        "MFh:SCOP",             "MFh:WID",          "MFh:STEP",
        "MFi:MTR",              "MFi:SWR",          "MFi:DISP",
        "MFj:SPOT",             "MFj:BK",           "MFj:KYR",
        "MFk:TUNE",             "MFk:DOWN",         "MFk:UP",
        "MFl:NB",               "MFl:AGC",          "MFl:AGC SEL",
        "MFm:IPO",              "MFm:ATT",          "MFm:NAR",
        "MFn:CFIL",             "MFn:FIL1",         "MFn:FIL2",
        "MFo:PLY1",             "MFo:PLY2",         "MFo:PLY3",
        "MFp:DNR",              "MFp:DNF",          "MFp:DBF",
        "01:EXT MENU",          "02:144MHz ARS",    "03:430MHz ARS",
        "04:AM&FM DIAL",        "05:AM MIC GAIN",   "06:AM STEP",
        "07:APO TIME",          "08:ARTS BEEP",     "09:ARTS ID",
        "10:ARTS IDW",          "11:BEACON TEXT",   "12:BEACON TIME",
        "13:BEEP TONE",         "14:BEEP VOL",      "15:CAR LSB R",
        "16:CAR LSB T",         "17:CAR USB R",     "18:CAR USB T",
        "19:CAT RATE",          "20:CAT/LIN/TUN",   "21:CLAR DIAL SEL",
        "22:CW AUTO MODE",      "23:CW BFO",        "24:CW DELAY",
        "25:CW KEY REV",        "26:CW PADDLE",     "27:CW PITCH",
        "28:CW QSK",            "29:CW SIDE TONE",  "30:CW SPEED",
        "31:CW TRAINING",       "32:CW WEIGHT",     "33:DCS CODE",
        "34:DCS INV",           "35:DIAL STEP",     "36:DIG DISP",
        "37:DIG GAIN",          "38:DIG MODE",      "39:DIG SHIFT",
        "40:DIG VOX",           "41:DISP COLOR",    "42:DISP CONTRAST",
        "43:DISP INTENSITY",    "44:DISP MODE",     "45:DSP BPF WIDTH",
        "46:DSP HPF CUTOFF",    "47:DSP LPF CUTOFF", "48:DSP MIC EQ",
        "49:DSP NR LEVEL",      "50:EMERGENCY",     "51:FM MIC GAIN",
        "52:FM STEP",           "53:HOME->VFO",     "54:LOCK MODE",
        "55:MEM GROUP",         "56:MEM TAG",       "57:MEM/VFO DIAL MODE",
        "58:MIC SCAN",          "59:MIC SEL",       "60:MTR ARX",
        "61:MTR ATX",           "62:MTR PEAK HOLD", "63:NB LEVEL",
        "64:OP FILTER",         "71:PKT 1200",      "72:PKT 9600",
        "73:PKT RATE",          "74:PROC LEVEL",    "75:RF POWER SET",
        "76:RPT SHIFT",         "77:SCAN MODE",     "78:SCAN RESUME",
        "79:SPLIT TONE",        "80:SQL/RF GAIN",   "81:SSB MIC GAIN",
        "82:SSB STEP",          "83:TONE FREQ",     "84:TX TIME",
        "85:TUNER/ATAS",        "86:TX IF FILTER",  "87:VOX DELAY",
        "88:VOX GAIN",          "89:XVTR A FREQ",   "90:XVTR B FREQ",
        "91:XVTR SEL",
        "MONI", "Q.SPL", "TCALL", "ATC", "USER"]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to CAT/LINEAR jack.\n"
            "3. Press and hold in the [MODE &lt;] and [MODE &gt;] keys"
            " while\n"
            "     turning the radio on (\"CLONE MODE\" will appear on the\n"
            "     display).\n"
            "4. <b>After clicking OK</b>,\n"
            "     press the [C](SEND) key to send image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect cable to ACC jack.\n"
            "3. Press and hold in the [MODE &lt;] and [MODE &gt;] keys"
            " while\n"
            "     turning the radio on (\"CLONE MODE\" will appear on the\n"
            "     display).\n"
            "4. Press the [A](RCV) key (\"receiving\" will appear on the"
            " LCD).\n")
        return rp

    def get_features(self):
        rf = ft817.FT817Radio.get_features(self)
        rf.has_cross = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.valid_tmodes = list(self.TMODES_REV.keys())
        rf.valid_cross_modes = list(self.CROSS_MODES_REV.keys())
        return rf

    def _get_duplex(self, mem, _mem):
        # radio set is_duplex only for + and - but not for split
        # at the same time it does not complain if we set it same way 817 does
        # (so no set_duplex here)
        mem.duplex = self.DUPLEX[_mem.duplex]

    def _get_tmode(self, mem, _mem):
        if not _mem.is_split_tone:
            mem.tmode = self.TMODES[int(_mem.tmode)]
        else:
            mem.tmode = "Cross"
            mem.cross_mode = self.CROSS_MODES[int(_mem.tmode)]

        if mem.tmode == "Tone":
            mem.rtone = mem.ctone = chirp_common.TONES[_mem.tone]
        elif mem.tmode == "TSQL":
            mem.rtone = mem.ctone = chirp_common.TONES[_mem.tone]
        elif mem.tmode == "DTCS Enc":   # UI does not support it yet but
            # this code has already been tested
            mem.dtcs = mem.rx_dtcs = chirp_common.DTCS_CODES[_mem.dcs]
        elif mem.tmode == "DTCS":
            mem.dtcs = mem.rx_dtcs = chirp_common.DTCS_CODES[_mem.dcs]
        elif mem.tmode == "Cross":
            mem.ctone = chirp_common.TONES[_mem.rxtone]
            # don't want to fail for this
            try:
                mem.rtone = chirp_common.TONES[_mem.tone]
            except IndexError:
                mem.rtone = chirp_common.TONES[0]
            mem.dtcs = chirp_common.DTCS_CODES[_mem.dcs]
            mem.rx_dtcs = chirp_common.DTCS_CODES[_mem.rxdcs]

    def _set_tmode(self, mem, _mem):
        if mem.tmode != "Cross":
            _mem.is_split_tone = 0
            _mem.tmode = self.TMODES_REV[mem.tmode]
        else:
            _mem.tmode = self.CROSS_MODES_REV[mem.cross_mode]
            _mem.is_split_tone = 1

        if mem.tmode == "Tone":
            _mem.tone = _mem.rxtone = chirp_common.TONES.index(mem.rtone)
        elif mem.tmode == "TSQL":
            _mem.tone = _mem.rxtone = chirp_common.TONES.index(mem.ctone)
        elif mem.tmode == "DTCS Enc":   # UI does not support it yet but
            # this code has already been tested
            _mem.dcs = _mem.rxdcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        elif mem.tmode == "DTCS":
            _mem.dcs = _mem.rxdcs = chirp_common.DTCS_CODES.index(mem.rx_dtcs)
        elif mem.tmode == "Cross":
            _mem.tone = chirp_common.TONES.index(mem.rtone)
            _mem.rxtone = chirp_common.TONES.index(mem.ctone)
            _mem.dcs = chirp_common.DTCS_CODES.index(mem.dtcs)
            _mem.rxdcs = chirp_common.DTCS_CODES.index(mem.rx_dtcs)
        # have to put this bit to 0 otherwise we get strange display in tone
        # frequency (menu 83). See bug #88 and #163
        _mem.unknown_toneflag = 0
        # dunno if there's the same problem here but to be safe ...
        _mem.unknown_rxtoneflag = 0

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic")
        cw = RadioSettingGroup("cw", "CW")
        packet = RadioSettingGroup("packet", "Digital & packet")
        panel = RadioSettingGroup("panel", "Panel settings")
        extended = RadioSettingGroup("extended", "Extended")
        panelcontr = RadioSettingGroup("panelcontr", "Panel controls")

        top = RadioSettings(basic, cw, packet,
                            panelcontr, panel, extended)

        rs = RadioSetting("extended_menu", "Extended menu",
                          RadioSettingValueBoolean(_settings.extended_menu))
        extended.append(rs)
        rs = RadioSetting("ars_144", "144 MHz ARS",
                          RadioSettingValueBoolean(_settings.ars_144))
        basic.append(rs)
        rs = RadioSetting("ars_430", "430 MHz ARS",
                          RadioSettingValueBoolean(_settings.ars_430))
        basic.append(rs)
        options = ["enable", "disable"]
        rs = RadioSetting("disable_amfm_dial", "AM&FM Dial",
                          RadioSettingValueList(options,
                                                options[
                                                    _settings.disable_amfm_dial
                                                        ]))
        panel.append(rs)
        rs = RadioSetting("am_mic", "AM mic gain",
                          RadioSettingValueInteger(0, 100, _settings.am_mic))
        basic.append(rs)
        options = ["OFF", "1h", "2h", "3h", "4h", "5h", "6h"]
        rs = RadioSetting("apo_time", "APO time",
                          RadioSettingValueList(options,
                                                options[_settings.apo_time]))
        basic.append(rs)
        options = ["OFF", "Range", "All"]
        rs = RadioSetting("arts_beep", "ARTS beep",
                          RadioSettingValueList(options,
                                                options[_settings.arts_beep]))
        basic.append(rs)
        rs = RadioSetting("arts_id", "ARTS ID",
                          RadioSettingValueBoolean(_settings.arts_id))
        extended.append(rs)
        st = RadioSettingValueString(0, 10,
                                     safe_charset_string(
                                         self._memobj.arts_idw,
                                         self._CALLSIGN_CHARSET)
                                     )
        st.set_charset(self._CALLSIGN_CHARSET)
        rs = RadioSetting("arts_idw", "ARTS IDW", st)
        extended.append(rs)
        st = RadioSettingValueString(0, 40,
                                     safe_charset_string(
                                         self._memobj.beacon_text1,
                                         self._BEACON_CHARSET)
                                     )
        st.set_charset(self._BEACON_CHARSET)
        rs = RadioSetting("beacon_text1", "Beacon text1", st)
        extended.append(rs)
        st = RadioSettingValueString(0, 40,
                                     safe_charset_string(
                                         self._memobj.beacon_text2,
                                         self._BEACON_CHARSET)
                                     )
        st.set_charset(self._BEACON_CHARSET)
        rs = RadioSetting("beacon_text2", "Beacon text2", st)
        extended.append(rs)
        st = RadioSettingValueString(0, 40,
                                     safe_charset_string(
                                         self._memobj.beacon_text3,
                                         self._BEACON_CHARSET)
                                     )
        st.set_charset(self._BEACON_CHARSET)
        rs = RadioSetting("beacon_text3", "Beacon text3", st)
        extended.append(rs)
        options = ["OFF"] + ["%i sec" % i for i in range(1, 256)]
        rs = RadioSetting("beacon_time", "Beacon time",
                          RadioSettingValueList(options,
                                                options[_settings.beacon_time])
                          )
        extended.append(rs)
        options = ["440 Hz", "880 Hz", "1760 Hz"]
        rs = RadioSetting("beep_tone", "Beep tone",
                          RadioSettingValueList(options,
                                                options[_settings.beep_tone]))
        panel.append(rs)
        rs = RadioSetting("beep_vol", "Beep volume",
                          RadioSettingValueInteger(0, 100, _settings.beep_vol))
        panel.append(rs)
        rs = RadioSetting("r_lsb_car", "LSB Rx carrier point (*10 Hz)",
                          RadioSettingValueInteger(-30, 30,
                                                   _settings.r_lsb_car))
        extended.append(rs)
        rs = RadioSetting("r_usb_car", "USB Rx carrier point (*10 Hz)",
                          RadioSettingValueInteger(-30, 30,
                                                   _settings.r_usb_car))
        extended.append(rs)
        rs = RadioSetting("t_lsb_car", "LSB Tx carrier point (*10 Hz)",
                          RadioSettingValueInteger(-30, 30,
                                                   _settings.t_lsb_car))
        extended.append(rs)
        rs = RadioSetting("t_usb_car", "USB Tx carrier point (*10 Hz)",
                          RadioSettingValueInteger(-30, 30,
                                                   _settings.t_usb_car))
        extended.append(rs)
        options = ["4800", "9600", "38400"]
        rs = RadioSetting("cat_rate", "CAT rate",
                          RadioSettingValueList(options,
                                                options[_settings.cat_rate]))
        basic.append(rs)
        options = ["CAT", "Linear", "Tuner"]
        rs = RadioSetting("cat_lin_tun", "CAT/LIN/TUN selection",
                          RadioSettingValueList(options,
                                                options[_settings.cat_lin_tun])
                          )
        extended.append(rs)
        options = ["MAIN", "VFO/MEM", "CLAR"]
        # TODO test the 3 options on non D radio
        # which have only SEL and MAIN
        rs = RadioSetting("clar_dial_sel", "Clarifier dial selection",
                          RadioSettingValueList(options,
                                                options[
                                                    _settings.clar_dial_sel]))
        panel.append(rs)
        rs = RadioSetting("cw_auto_mode", "CW Automatic mode",
                          RadioSettingValueBoolean(_settings.cw_auto_mode))
        cw.append(rs)
        options = ["USB", "LSB", "AUTO"]
        rs = RadioSetting("cw_bfo", "CW BFO",
                          RadioSettingValueList(options,
                                                options[_settings.cw_bfo]))
        cw.append(rs)
        options = ["FULL"] + ["%i ms" % (i * 10) for i in range(3, 301)]
        val = (_settings.cw_delay + _settings.cw_delay_hi * 256) - 2
        rs = RadioSetting("cw_delay", "CW delay",
                          RadioSettingValueList(options, options[val]))
        cw.append(rs)
        options = ["Normal", "Reverse"]
        rs = RadioSetting("cw_key_rev", "CW key reverse",
                          RadioSettingValueList(options,
                                                options[_settings.cw_key_rev]))
        cw.append(rs)
        rs = RadioSetting("cw_paddle", "CW paddle",
                          RadioSettingValueBoolean(_settings.cw_paddle))
        cw.append(rs)
        options = ["%i Hz" % i for i in range(400, 801, 100)]
        rs = RadioSetting("cw_pitch", "CW pitch",
                          RadioSettingValueList(options,
                                                options[_settings.cw_pitch]))
        cw.append(rs)
        options = ["%i ms" % i for i in range(5, 31, 5)]
        rs = RadioSetting("cw_qsk", "CW QSK",
                          RadioSettingValueList(options,
                                                options[_settings.cw_qsk]))
        cw.append(rs)
        rs = RadioSetting("cw_sidetone", "CW sidetone volume",
                          RadioSettingValueInteger(0, 100,
                                                   _settings.cw_sidetone))
        cw.append(rs)
        options = ["%i wpm" % i for i in range(4, 61)]
        rs = RadioSetting("cw_speed", "CW speed",
                          RadioSettingValueList(options,
                                                options[_settings.cw_speed]))
        cw.append(rs)
        options = ["Numeric", "Alphabet", "AlphaNumeric"]
        rs = RadioSetting("cw_training", "CW training",
                          RadioSettingValueList(options,
                                                options[_settings.cw_training])
                          )
        cw.append(rs)
        options = ["1:%1.1f" % (i / 10) for i in range(25, 46, 1)]
        rs = RadioSetting("cw_weight", "CW weight",
                          RadioSettingValueList(options,
                                                options[_settings.cw_weight]))
        cw.append(rs)
        options = ["Tn-Rn", "Tn-Riv", "Tiv-Rn", "Tiv-Riv"]
        rs = RadioSetting("dcs_inv", "DCS inv",
                          RadioSettingValueList(options,
                                                options[_settings.dcs_inv]))
        extended.append(rs)
        options = ["Fine", "Coarse"]
        rs = RadioSetting("dial_step", "Dial step",
                          RadioSettingValueList(options,
                                                options[_settings.dial_step]))
        panel.append(rs)
        rs = RadioSetting("dig_disp", "Dig disp (*10 Hz)",
                          RadioSettingValueInteger(-300, 300,
                                                   _settings.dig_disp))
        packet.append(rs)
        rs = RadioSetting("dig_mic", "Dig gain",
                          RadioSettingValueInteger(0, 100,
                                                   _settings.dig_mic))
        packet.append(rs)
        options = ["RTTYL", "RTTYU", "PSK31-L", "PSK31-U", "USER-L", "USER-U"]
        rs = RadioSetting("dig_mode", "Dig mode",
                          RadioSettingValueList(options,
                                                options[_settings.dig_mode]))
        packet.append(rs)
        rs = RadioSetting("dig_shift", "Dig shift (*10 Hz)",
                          RadioSettingValueInteger(-300, 300,
                                                   _settings.dig_shift))
        packet.append(rs)
        rs = RadioSetting("dig_vox", "Dig vox",
                          RadioSettingValueInteger(0, 100, _settings.dig_vox))
        packet.append(rs)
        options = ["ARTS", "BAND", "FIX", "MEMGRP", "MODE", "MTR", "VFO"]
        rs = RadioSetting("disp_color", "Display color mode",
                          RadioSettingValueList(options,
                                                options[_settings.disp_color]))
        panel.append(rs)
        rs = RadioSetting("disp_color_arts", "Display color ARTS set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_arts))
        panel.append(rs)
        rs = RadioSetting("disp_color_band", "Display color band set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_band))
        panel.append(rs)
        rs = RadioSetting("disp_color_memgrp",
                          "Display color memory group set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_memgrp)
                          )
        panel.append(rs)
        rs = RadioSetting("disp_color_mode", "Display color mode set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_mode))
        panel.append(rs)
        rs = RadioSetting("disp_color_mtr", "Display color meter set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_mtr))
        panel.append(rs)
        rs = RadioSetting("disp_color_vfo", "Display color VFO set",
                          RadioSettingValueInteger(0, 1,
                                                   _settings.disp_color_vfo))
        panel.append(rs)
        rs = RadioSetting("disp_color_fix", "Display color fix set",
                          RadioSettingValueInteger(1, 32,
                                                   _settings.disp_color_fix + 1
                                                   ))
        panel.append(rs)
        rs = RadioSetting("disp_contrast", "Contrast",
                          RadioSettingValueInteger(3, 15,
                                                   _settings.disp_contrast + 2)
                          )
        panel.append(rs)
        rs = RadioSetting("disp_intensity", "Intensity",
                          RadioSettingValueInteger(1, 3,
                                                   _settings.disp_intensity))
        panel.append(rs)
        options = ["OFF", "Auto1", "Auto2", "ON"]
        rs = RadioSetting("disp_mode", "Display backlight mode",
                          RadioSettingValueList(options,
                                                options[_settings.disp_mode]))
        panel.append(rs)
        options = ["60 Hz", "120 Hz", "240 Hz"]
        rs = RadioSetting("dsp_bpf", "Dsp band pass filter",
                          RadioSettingValueList(options,
                                                options[_settings.dsp_bpf]))
        cw.append(rs)
        options = ["100 Hz", "160 Hz", "220 Hz", "280 Hz", "340 Hz", "400 Hz",
                   "460 Hz", "520 Hz", "580 Hz", "640 Hz", "720 Hz", "760 Hz",
                   "820 Hz", "880 Hz", "940 Hz", "1000 Hz"]
        rs = RadioSetting("dsp_hpf", "Dsp hi pass filter cut off",
                          RadioSettingValueList(options,
                                                options[_settings.dsp_hpf]))
        basic.append(rs)
        options = ["1000 Hz", "1160 Hz", "1320 Hz", "1480 Hz", "1650 Hz",
                   "1800 Hz", "1970 Hz", "2130 Hz", "2290 Hz", "2450 Hz",
                   "2610 Hz", "2770 Hz", "2940 Hz", "3100 Hz", "3260 Hz",
                   "3420 Hz", "3580 Hz", "3740 Hz", "3900 Hz", "4060 Hz",
                   "4230 Hz", "4390 Hz", "4550 Hz", "4710 Hz", "4870 Hz",
                   "5030 Hz", "5190 Hz", "5390 Hz", "5520 Hz", "5680 Hz",
                   "5840 Hz", "6000 Hz"]
        rs = RadioSetting("dsp_lpf", "Dsp low pass filter cut off",
                          RadioSettingValueList(options,
                                                options[_settings.dsp_lpf]))
        basic.append(rs)
        options = ["OFF", "LPF", "HPF", "BOTH"]
        rs = RadioSetting("dsp_mic_eq", "Dsp mic equalization",
                          RadioSettingValueList(options,
                                                options[_settings.dsp_mic_eq]))
        basic.append(rs)
        rs = RadioSetting("dsp_nr", "DSP noise reduction level",
                          RadioSettingValueInteger(1, 16,
                                                   _settings.dsp_nr + 1))
        basic.append(rs)
        # emergency only for US model
        rs = RadioSetting("fm_mic", "FM mic gain",
                          RadioSettingValueInteger(0, 100, _settings.fm_mic))
        basic.append(rs)
        rs = RadioSetting("home_vfo", "Enable HOME to VFO moving",
                          RadioSettingValueBoolean(_settings.home_vfo))
        panel.append(rs)
        options = ["Dial", "Freq", "Panel", "All"]
        rs = RadioSetting("lock_mode", "Lock mode",
                          RadioSettingValueList(options,
                                                options[_settings.lock_mode]))
        panel.append(rs)
        rs = RadioSetting("mem_group", "Mem group",
                          RadioSettingValueBoolean(_settings.mem_group))
        basic.append(rs)
        options = ["CW SIDETONE", "CW SPEED", "MHz/MEM GRP", "MIC GAIN",
                   "NB LEVEL", "RF POWER", "STEP"]
        rs = RadioSetting("mem_vfo_dial_mode", "Mem/VFO dial mode",
                          RadioSettingValueList(options,
                                                options[
                                                    _settings.mem_vfo_dial_mode
                                                       ]))
        panel.append(rs)
        rs = RadioSetting("mic_scan", "Mic scan",
                          RadioSettingValueBoolean(_settings.mic_scan))
        basic.append(rs)
        options = ["NOR", "RMT", "CAT"]
        rs = RadioSetting("mic_sel", "Mic selection",
                          RadioSettingValueList(options,
                                                options[_settings.mic_sel]))
        extended.append(rs)
        options = ["SIG", "CTR", "VLT", "N/A", "FS", "OFF"]
        rs = RadioSetting("mtr_arx_sel", "Meter receive selection",
                          RadioSettingValueList(options,
                                                options[_settings.mtr_arx_sel])
                          )
        extended.append(rs)
        options = ["PWR", "ALC", "MOD", "SWR", "VLT", "N/A", "OFF"]
        rs = RadioSetting("mtr_atx_sel", "Meter transmit selection",
                          RadioSettingValueList(options,
                                                options[_settings.mtr_atx_sel])
                          )
        extended.append(rs)
        rs = RadioSetting("mtr_peak_hold", "Meter peak hold",
                          RadioSettingValueBoolean(_settings.mtr_peak_hold))
        extended.append(rs)
        rs = RadioSetting("nb_level", "Noise blanking level",
                          RadioSettingValueInteger(0, 100, _settings.nb_level))
        basic.append(rs)
        st = RadioSettingValueString(0, 4,
                                     safe_charset_string(
                                         self._memobj.op_filter1_name,
                                         self._CALLSIGN_CHARSET)
                                     )
        st.set_charset(self._CALLSIGN_CHARSET)
        rs = RadioSetting("op_filter1_name", "Optional filter1 name", st)
        extended.append(rs)
        st = RadioSettingValueString(0, 4,
                                     safe_charset_string(
                                         self._memobj.op_filter2_name,
                                         self._CALLSIGN_CHARSET)
                                     )
        st.set_charset(self._CALLSIGN_CHARSET)
        rs = RadioSetting("op_filter2_name", "Optional filter2 name", st)
        extended.append(rs)
        rs = RadioSetting("pg_a", "Programmable key MFq:A",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_a]))
        extended.append(rs)
        rs = RadioSetting("pg_b", "Programmable key MFq:B",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_b]))
        extended.append(rs)
        rs = RadioSetting("pg_c", "Programmable key MFq:C",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_c]))
        extended.append(rs)
        rs = RadioSetting("pg_acc", "Programmable mic key ACC",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_acc]))
        extended.append(rs)
        rs = RadioSetting("pg_p1", "Programmable mic key P1",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_p1]))
        extended.append(rs)
        rs = RadioSetting("pg_p2", "Programmable mic key P2",
                          RadioSettingValueList(self.PROGRAMMABLEOPTIONS,
                                                self.PROGRAMMABLEOPTIONS[
                                                    _settings.pg_p2]))
        extended.append(rs)
        rs = RadioSetting("pkt1200", "Packet 1200 gain level",
                          RadioSettingValueInteger(0, 100,
                                                   _settings.pkt1200))
        packet.append(rs)
        rs = RadioSetting("pkt9600", "Packet 9600 gain level",
                          RadioSettingValueInteger(0, 100, _settings.pkt9600))
        packet.append(rs)
        options = ["1200", "9600"]
        rs = RadioSetting("pkt_rate", "Packet rate",
                          RadioSettingValueList(options,
                                                options[_settings.pkt_rate]))
        packet.append(rs)
        rs = RadioSetting("proc_level", "Proc level",
                          RadioSettingValueInteger(0, 100,
                                                   _settings.proc_level))
        basic.append(rs)
        rs = RadioSetting("rf_power_hf", "Rf power set HF",
                          RadioSettingValueInteger(5, 100,
                                                   _settings.rf_power_hf))
        basic.append(rs)
        rs = RadioSetting("rf_power_6m", "Rf power set 6m",
                          RadioSettingValueInteger(5, 100,
                                                   _settings.rf_power_6m))
        basic.append(rs)
        rs = RadioSetting("rf_power_vhf", "Rf power set VHF",
                          RadioSettingValueInteger(5, 50,
                                                   _settings.rf_power_vhf))
        basic.append(rs)
        rs = RadioSetting("rf_power_uhf", "Rf power set UHF",
                          RadioSettingValueInteger(2, 20,
                                                   _settings.rf_power_uhf))
        basic.append(rs)
        options = ["TIME", "BUSY", "STOP"]
        rs = RadioSetting("scan_mode", "Scan mode",
                          RadioSettingValueList(options,
                                                options[_settings.scan_mode]))
        basic.append(rs)
        rs = RadioSetting("scan_resume", "Scan resume",
                          RadioSettingValueInteger(1, 10,
                                                   _settings.scan_resume))
        basic.append(rs)
        rs = RadioSetting("split_tone", "Split tone enable",
                          RadioSettingValueBoolean(_settings.split_tone))
        extended.append(rs)
        options = ["RF-Gain", "Squelch"]
        rs = RadioSetting("sql_rf_gain", "Squelch/RF-Gain",
                          RadioSettingValueList(options,
                                                options[_settings.sql_rf_gain])
                          )
        panel.append(rs)
        rs = RadioSetting("ssb_mic", "SSB Mic gain",
                          RadioSettingValueInteger(0, 100, _settings.ssb_mic))
        basic.append(rs)
        options = ["Off"] + ["%i" % i for i in range(1, 21)]
        rs = RadioSetting("tot_time", "Time-out timer",
                          RadioSettingValueList(options,
                                                options[_settings.tot_time]))
        basic.append(rs)
        options = ["OFF", "ATAS(HF)", "ATAS(HF&50)", "ATAS(ALL)", "TUNER"]
        rs = RadioSetting("tuner_atas", "Tuner/ATAS device",
                          RadioSettingValueList(options,
                                                options[_settings.tuner_atas]))
        extended.append(rs)
        rs = RadioSetting("tx_if_filter", "Transmit IF filter",
                          RadioSettingValueList(self.FILTERS,
                                                self.FILTERS[
                                                    _settings.tx_if_filter]))
        basic.append(rs)
        rs = RadioSetting("vox_delay", "VOX delay (*100 ms)",
                          RadioSettingValueInteger(1, 30, _settings.vox_delay))
        basic.append(rs)
        rs = RadioSetting("vox_gain", "VOX Gain",
                          RadioSettingValueInteger(1, 100, _settings.vox_gain))
        basic.append(rs)
        rs = RadioSetting("xvtr_a", "Xvtr A displacement",
                          RadioSettingValueInteger(
                              -4294967295, 4294967295,
                              self._memobj.xvtr_a_offset *
                              (-1 if _settings.xvtr_a_negative else 1)))

        extended.append(rs)
        rs = RadioSetting("xvtr_b", "Xvtr B displacement",
                          RadioSettingValueInteger(
                              -4294967295, 4294967295,
                              self._memobj.xvtr_b_offset *
                              (-1 if _settings.xvtr_b_negative else 1)))
        extended.append(rs)
        options = ["OFF", "XVTR A", "XVTR B"]
        rs = RadioSetting("xvtr_sel", "Transverter function selection",
                          RadioSettingValueList(options,
                                                options[_settings.xvtr_sel]))
        extended.append(rs)

        rs = RadioSetting("disp", "Display large",
                          RadioSettingValueBoolean(_settings.disp))
        panel.append(rs)
        rs = RadioSetting("nb", "Noise blanker",
                          RadioSettingValueBoolean(_settings.nb))
        panelcontr.append(rs)
        options = ["Auto", "Fast", "Slow", "Off"]
        rs = RadioSetting("agc", "AGC",
                          RadioSettingValueList(options,
                                                options[_settings.agc]))
        panelcontr.append(rs)
        options = ["PWR", "ALC", "SWR", "MOD"]
        rs = RadioSetting("pwr_meter_mode", "Power meter mode",
                          RadioSettingValueList(options,
                                                options[
                                                    _settings.pwr_meter_mode]))
        panelcontr.append(rs)
        rs = RadioSetting("vox", "Vox",
                          RadioSettingValueBoolean(_settings.vox))
        panelcontr.append(rs)
        rs = RadioSetting("bk", "Semi break-in",
                          RadioSettingValueBoolean(_settings.bk))
        cw.append(rs)
        rs = RadioSetting("kyr", "Keyer",
                          RadioSettingValueBoolean(_settings.kyr))
        cw.append(rs)
        options = ["enabled", "disabled"]
        rs = RadioSetting("fst", "Fast",
                          RadioSettingValueList(options, options[_settings.fst]
                                                ))
        panelcontr.append(rs)
        options = ["enabled", "disabled"]
        rs = RadioSetting("lock", "Lock",
                          RadioSettingValueList(options,
                                                options[_settings.lock]))
        panelcontr.append(rs)
        rs = RadioSetting("scope_peakhold", "Scope max hold",
                          RadioSettingValueBoolean(_settings.scope_peakhold))
        panelcontr.append(rs)
        options = ["21", "31", "127"]
        rs = RadioSetting("scope_width", "Scope width (channels)",
                          RadioSettingValueList(options,
                                                options[_settings.scope_width])
                          )
        panelcontr.append(rs)
        rs = RadioSetting("proc", "Speech processor",
                          RadioSettingValueBoolean(_settings.proc))
        panelcontr.append(rs)

        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            try:
                if "." in element.get_name():
                    bits = element.get_name().split(".")
                    obj = self._memobj
                    for bit in bits[:-1]:
                        obj = getattr(obj, bit)
                    setting = bits[-1]
                else:
                    obj = _settings
                    setting = element.get_name()
                try:
                    LOG.debug("Setting %s(%s) <= %s" % (setting,
                              getattr(obj, setting), element.value))
                except AttributeError:
                    LOG.debug("Setting %s <= %s" % (setting, element.value))
                if setting == "arts_idw":
                    self._memobj.arts_idw = \
                        [self._CALLSIGN_CHARSET_REV[x]
                         for x in str(element.value)]
                elif setting in ["beacon_text1", "beacon_text2",
                                 "beacon_text3", "op_filter1_name",
                                 "op_filter2_name"]:
                    setattr(self._memobj, setting,
                            [self._BEACON_CHARSET_REV[x]
                             for x in str(element.value)])
                elif setting == "cw_delay":
                    val = int(element.value) + 2
                    setattr(obj, "cw_delay_hi", val / 256)
                    setattr(obj, setting, val & 0xff)
                elif setting == "dig_vox":
                    val = int(element.value)
                    setattr(obj, "dig_vox_enable", int(val > 0))
                    setattr(obj, setting, val)
                elif setting in ["disp_color_fix", "dsp_nr"]:
                    setattr(obj, setting, int(element.value) - 1)
                elif setting == "disp_contrast":
                    setattr(obj, setting, int(element.value) - 2)
                elif setting in ["xvtr_a", "xvtr_b"]:
                    val = int(element.value)
                    setattr(obj, setting + "_negative", int(val < 0))
                    setattr(self._memobj, setting + "_offset", abs(val))
                else:
                    setattr(obj, setting, element.value)
            except Exception:
                LOG.debug(element.get_name())
                raise


@directory.register
class FT857USRadio(FT857Radio):
    """Yaesu FT857/897 (US version)"""
    # seems that radios configured for 5 MHz operations send one packet more
    # than others so we have to distinguish sub models
    MODEL = "FT-857/897 (US)"

    _model = ""
    _US_model = True
    _memsize = 7481
    # block 9 (140 Bytes long) is to be repeated 40 times
    # should be 42 times but this way I can use original 817 functions
    _block_lengths = [2, 82, 252, 196, 252, 196, 212, 55, 140, 140, 140, 38,
                      176, 140]

    SPECIAL_60M = {
        "M-601": -52,
        "M-602": -51,
        "M-603": -50,
        "M-604": -49,
        "M-605": -48,
        }
    LAST_SPECIAL60M_INDEX = -52

    SPECIAL_MEMORIES = dict(FT857Radio.SPECIAL_MEMORIES)
    SPECIAL_MEMORIES.update(SPECIAL_60M)

    SPECIAL_MEMORIES_REV = dict(list(zip(list(SPECIAL_MEMORIES.values()),
                                         list(SPECIAL_MEMORIES.keys()))))

    # this is identical to the one in FT817ND_US_Radio but we inherit from 857
    def _get_special_60m(self, number):
        mem = chirp_common.Memory()
        mem.number = self.SPECIAL_60M[number]
        mem.extd_number = number

        _mem = self._memobj.sixtymeterchannels[-self.LAST_SPECIAL60M_INDEX +
                                               mem.number]

        mem = self._get_memory(mem, _mem)

        mem.immutable = ["number", "rtone", "ctone",
                         "extd_number", "name", "dtcs", "tmode", "cross_mode",
                         "dtcs_polarity", "power", "duplex", "offset",
                         "empty"]

        return mem

    # this is identical to the one in FT817ND_US_Radio but we inherit from 857
    def _set_special_60m(self, mem):
        if mem.empty:
            # can't delete 60M memories!
            raise Exception("Sorry, 60M memory can't be deleted")

        cur_mem = self._get_special_60m(self.SPECIAL_MEMORIES_REV[mem.number])

        for key in cur_mem.immutable:
            if cur_mem.__dict__[key] != mem.__dict__[key]:
                raise errors.RadioError("Editing field `%s' " % key +
                                        "is not supported on M-60x channels")

        if mem.mode not in ["USB", "LSB", "CW", "CWR", "NCW", "NCWR", "DIG"]:
            raise errors.RadioError("Mode {mode} is not valid "
                                    "in 60m channels".format(mode=mem.mode))
        _mem = self._memobj.sixtymeterchannels[-self.LAST_SPECIAL60M_INDEX +
                                               mem.number]
        self._set_memory(mem, _mem)

    def get_memory(self, number):
        if number in self.SPECIAL_60M:
            return self._get_special_60m(number)
        elif isinstance(number, int) and number < 0 and (
                self.SPECIAL_MEMORIES_REV[number] in
                list(self.SPECIAL_60M.keys())):
            # I can't stop delete operation from losing extd_number but
            # I know how to get it back
            return self._get_special_60m(self.SPECIAL_MEMORIES_REV[number])
        else:
            return FT857Radio.get_memory(self, number)

    def set_memory(self, memory):
        if memory.number in list(self.SPECIAL_60M.values()):
            return self._set_special_60m(memory)
        else:
            return FT857Radio.set_memory(self, memory)

    def get_settings(self):
        top = FT857Radio.get_settings(self)
        basic = top[0]
        rs = RadioSetting("emergency", "Emergency",
                          RadioSettingValueBoolean(
                              self._memobj.settings.emergency))
        basic.append(rs)
        return top



# Copyright 2022 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct memory {
  u32 rxfreq;                                // 00-03
  u16 decQT;                                 // 04-05
  u32 txfreq;                                // 06-09
  u16 encQT;                                 // 0a-0b
  u8 lowpower:1,  // Power Level             // 0c
     unknown1:1,
     isnarrow:1,  // Bandwidth
     bcl:2,       // Busy Channel Lockout
     scan:1,      // Scan Add
     encode:1,    // Encode
     isunused:1;  // Is Unused
  u8 unknown3[3];                            // 0d-0f
};

#seekto 0x0170;
struct memory channels[99];

#seekto 0x0162;
struct {
  u8 unknown_1:1,           // 0x0162
     voice:2,               //               Voice Prompt
     beep:1,                //               Beep Switch
     unknown_2:1,
     vox:1,                 //               VOX
     autolock:1,            //               Auto Lock
     vibrate:1;             //               Vibrate Switch
  u8 squelch:4,             // 0x0163        SQ Level
     unknown_3:1,
     volume:3;              //               Volume Level
  u8 voxl:4,                // 0x0164        VOX Level
     voxd:4;                //               VOX Delay
  u8 unknown_5:1,           // 0x0165
     save:3,                //               Power Save
     calltone:4;            //               Call Tone
  u8 unknown_6:4,           // 0x0166
     roger:2,               //               Roger Tone
     backlight:2;           //               Backlight Set
  u16 tot;                  // 0x0167-0x0168 Time-out Timer
  u8 unknown_7[3];          // 0x0169-0x016B
  u8 skeyul;                // 0x016C        Side Key Up Long
  u8 skeyus;                // 0x016D        Side Key Up Short
  u8 skeydl;                // 0x016E        Side Key Down Long
  u8 skeyds;                // 0x016F        Side Key Down Short
} settings;
"""

CMD_ACK = b"\x06"

RB15_DTCS = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

LIST_BACKLIGHT = ["Off", "On", "Auto"]
LIST_BCL = ["None", "Carrier", "QT/DQT Match"]
LIST_ROGER = ["Off", "Start", "End", "Start and End"]
LIST_SAVE = ["Off", "1:1", "1:2", "1:3", "1:4", "1:5"]
_STEP_LIST = [2.5, 5., 6.25, 10., 12.5, 20., 25., 50.]
LIST_VOICE = ["Off", "Chinese", "English"]
LIST_VOXD = ["0.0", "0.5", "1.0", "1.5", "2.0", "2.5", "3.0", "3.5", "4.0",
             "4.5", "5.0S"]

SKEY_CHOICES = ["None", "Scan", "Monitor", "VOX On/Off",
                "Local Alarm", "Remote Alarm", "Backlight On/Off", "Call Tone"]
SKEY_VALUES = [0x00, 0x01, 0x03, 0x04, 0x09, 0x0A, 0x13, 0x14]

TOT_CHOICES = ["Off", "15", "30", "45", "60", "75", "90", "105", "120",
               "135", "150", "165", "180", "195", "210", "225", "240",
               "255", "270", "285", "300", "315", "330", "345", "360",
               "375", "390", "405", "420", "435", "450", "465", "480",
               "495", "510", "525", "540", "555", "570", "585", "600"
               ]
TOT_VALUES = [0x00, 0x0F, 0x1E, 0x2D, 0x3C, 0x4B, 0x5A, 0x69, 0x78,
              0x87, 0x96, 0xA5, 0xB4, 0xC3, 0xD2, 0xE1, 0xF0,
              0xFF, 0x10E, 0x11D, 0x12C, 0x13B, 0x14A, 0x159, 0x168,
              0x177, 0x186, 0x195, 0x1A4, 0x1B3, 0x1C2, 0x1D1, 0x1E0,
              0x1EF, 0x1FE, 0x20D, 0x21C, 0x22B, 0x23A, 0x249, 0x258
              ]


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def tone2short(t):
    """Convert a string tone or DCS to an encoded u16
    """
    tone = str(t)
    if tone == "----":
        u16tone = 0x0000
    elif tone[0] == 'D':  # This is a DCS code
        c = tone[1: -1]
        code = int(c, 8)
        if tone[-1] == 'I':
            code |= 0x4000
        u16tone = code | 0x8000
    else:  # This is an analog CTCSS
        u16tone = int(tone[0:-2]+tone[-1]) & 0xffff  # strip the '.'
    return u16tone


def short2tone(tone):
    """ Map a binary CTCSS/DCS to a string name for the tone
    """
    if tone == 0xC000 or tone == 0xffff:
        ret = "----"
    else:
        code = tone & 0x3fff
        if tone & 0x4000:      # This is a DCS
            ret = "D%0.3oN" % code
        elif tone & 0x8000:  # This is an inverse code
            ret = "D%0.3oI" % code
        else:   # Just plain old analog CTCSS
            ret = "%4.1f" % (code / 10.0)
    return ret


def _rb15_enter_programming_mode(radio):
    serial = radio.pipe

    # lengthen the timeout here as these radios are resetting due to timeout
    radio.pipe.timeout = 0.75

    exito = False
    for i in range(0, 5):
        serial.write(radio.magic)
        ack = serial.read(1)

        try:
            if ack == CMD_ACK:
                exito = True
                break
        except:
            LOG.debug("Attempt #%s, failed, trying again" % i)
            pass

    # return timeout to default value
    radio.pipe.timeout = 0.25

    # check if we had EXITO
    if exito is False:
        msg = "The radio did not accept program mode after five tries.\n"
        msg += "Check you interface cable and power cycle your radio."
        raise errors.RadioError(msg)


def _rb15_exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(b"21" + b"\x05\xEE" + b"V")
    except:
        raise errors.RadioError("Radio refused to exit programming mode")


def _rb15_read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">BH", ord(b'R'), block_addr)

    ccs = bytes([_checksum(cmd)])

    expectedresponse = b"R" + cmd[1:]

    cmd = cmd + ccs

    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        response = serial.read(3 + block_size + 1)

        cs = bytes([_checksum(response[:-1])])

        if response[:3] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        chunk = response[3:]

        if chunk[-1:] != cs:
            raise Exception("Block failed checksum!")

        block_data = chunk[:-1]
    except:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    return block_data


def _rb15_write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">BH", ord(b'W'), block_addr)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    cs = bytes([_checksum(cmd + data)])
    data += cs

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _rb15_enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0x0000, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _rb15_read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _rb15_exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _rb15_enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE):
            status.cur = addr + radio.BLOCK_SIZE
            radio.status_fn(status)
            _rb15_write_block(radio, addr, radio.BLOCK_SIZE)

    _rb15_exit_programming_mode(radio)


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # if the two freqs are on the same Band this is not a split
            return False

    # if you get here is because the freq pairs are split
    return True


class RB15RadioBase(chirp_common.CloneModeRadio):
    """RETEVIS RB15 BASE"""
    VENDOR = "Retevis"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    BLOCK_SIZE = 0x10
    magic = b"21" + b"\x05\x10" + b"x"

    VALID_BANDS = [(400000000, 480000000)]

    _ranges = [
               (0x0150, 0x07A0),
              ]
    _memsize = 0x07A0

    _frs = _pmr = False

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = False
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_dtcs_codes = RB15_DTCS
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = _STEP_LIST
        rf.valid_bands = self.VALID_BANDS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = do_download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = data
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            do_upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _get_tone(self, _mem, mem):
        """Decode both the encode and decode CTSS/DCS codes from
        the memory channel and stuff them into the UI
        memory channel row.
        """
        txtone = short2tone(_mem.encQT)
        rxtone = short2tone(_mem.decQT)
        pt = "N"
        pr = "N"

        if txtone == "----":
            txmode = ""
        elif txtone[0] == "D":
            mem.dtcs = int(txtone[1:4])
            if txtone[4] == "I":
                pt = "R"
            txmode = "DTCS"
        else:
            mem.rtone = float(txtone)
            txmode = "Tone"

        if rxtone == "----":
            rxmode = ""
        elif rxtone[0] == "D":
            mem.rx_dtcs = int(rxtone[1:4])
            if rxtone[4] == "I":
                pr = "R"
            rxmode = "DTCS"
        else:
            mem.ctone = float(rxtone)
            rxmode = "Tone"

        if txmode == "Tone" and len(rxmode) == 0:
            mem.tmode = "Tone"
        elif (txmode == rxmode and txmode == "Tone" and
              mem.rtone == mem.ctone):
            mem.tmode = "TSQL"
        elif (txmode == rxmode and txmode == "DTCS" and
              mem.dtcs == mem.rx_dtcs):
            mem.tmode = "DTCS"
        elif (len(rxmode) + len(txmode)) > 0:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        mem.dtcs_polarity = pt + pr

        LOG.debug("_get_tone: Got TX %s (%i) RX %s (%i)" %
                  (txmode, _mem.encQT, rxmode, _mem.decQT))

    def _set_tone(self, mem, _mem):
        """Update the memory channel block CTCC/DCS tones
        from the UI fields
        """
        def _set_dcs(code, pol):
            val = int("%i" % code, 8) | 0x4000
            if pol == "R":
                val = int("%i" % code, 8) | 0x8000
            return val

        rx_mode = tx_mode = None
        rxtone = txtone = 0xC000

        if mem.tmode == "Tone":
            tx_mode = "Tone"
            txtone = int(mem.rtone * 10)
        elif mem.tmode == "TSQL":
            rx_mode = tx_mode = "Tone"
            rxtone = txtone = int(mem.ctone * 10)
        elif mem.tmode == "DTCS":
            tx_mode = rx_mode = "DTCS"
            txtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            rxtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[1])
        elif mem.tmode == "Cross":
            tx_mode, rx_mode = mem.cross_mode.split("->")
            if tx_mode == "DTCS":
                txtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            elif tx_mode == "Tone":
                txtone = int(mem.rtone * 10)
            if rx_mode == "DTCS":
                rxtone = _set_dcs(mem.rx_dtcs, mem.dtcs_polarity[1])
            elif rx_mode == "Tone":
                rxtone = int(mem.ctone * 10)

        _mem.decQT = rxtone
        _mem.encQT = txtone

        LOG.debug("Set TX %s (%i) RX %s (%i)" %
                  (tx_mode, _mem.encQT, rx_mode, _mem.decQT))

    def get_memory(self, number):
        mem = chirp_common.Memory()
        _mem = self._memobj.channels[number - 1]
        mem.number = number

        mem.freq = int(_mem.rxfreq) * 10

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if _mem.get_raw() == (b"\xFF" * 16):
            LOG.debug("Initializing empty memory")
            _mem.set_raw(b"\x00" * 16)

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # tx freq can be blank
        if _mem.txfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX freq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset != 0:
                if _split(self.get_features(), mem.freq, int(
                          _mem.txfreq) * 10):
                    mem.duplex = "split"
                    mem.offset = int(_mem.txfreq) * 10
                elif offset < 0:
                    mem.offset = abs(offset)
                    mem.duplex = "-"
                elif offset > 0:
                    mem.offset = offset
                    mem.duplex = "+"
            else:
                mem.offset = 0

        mem.mode = _mem.isnarrow and "NFM" or "FM"

        self._get_tone(_mem, mem)

        mem.power = self.POWER_LEVELS[_mem.lowpower]

        if not _mem.scan:
            mem.skip = "S"

        mem.extra = RadioSettingGroup("Extra", "extra")

        if _mem.bcl > 0x02:
            val = 0
        else:
            val = _mem.bcl
        rs = RadioSetting("bcl", "BCL",
                          RadioSettingValueList(
                              LIST_BCL, LIST_BCL[val]))
        mem.extra.append(rs)

        rs = RadioSetting("encode", "Encode",
                          RadioSettingValueBoolean(_mem.encode))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        LOG.debug("Setting %i(%s)" % (mem.number, mem.extd_number))
        _mem = self._memobj.channels[mem.number - 1]

        # if empty memory
        if mem.empty:
            _mem.set_raw("\xFF" * 16)
            return

        _mem.isunused = False
        _mem.unknown1 = False

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            _mem.txfreq.fill_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _mem.scan = mem.skip != "S"
        _mem.isnarrow = mem.mode == "NFM"

        self._set_tone(mem, _mem)

        _mem.lowpower = mem.power == self.POWER_LEVELS[1]

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        sidekey = RadioSettingGroup("sidekey", "Side Key Settings")
        voxset = RadioSettingGroup("vox", "VOX Settings")
        top = RadioSettings(basic, sidekey, voxset)

        voice = RadioSetting("voice", "Language", RadioSettingValueList(
                             LIST_VOICE, LIST_VOICE[_settings.voice]))
        basic.append(voice)

        beep = RadioSetting("beep", "Key Beep",
                            RadioSettingValueBoolean(_settings.beep))
        basic.append(beep)

        volume = RadioSetting("volume", "Volume Level",
                              RadioSettingValueInteger(
                                  0, 7, _settings.volume))
        basic.append(volume)

        save = RadioSetting("save", "Battery Save",
                            RadioSettingValueList(
                                LIST_SAVE, LIST_SAVE[_settings.save]))
        basic.append(save)

        backlight = RadioSetting("backlight", "Backlight",
                                 RadioSettingValueList(
                                     LIST_BACKLIGHT,
                                     LIST_BACKLIGHT[_settings.backlight]))
        basic.append(backlight)

        vibrate = RadioSetting("vibrate", "Vibrate",
                               RadioSettingValueBoolean(_settings.vibrate))
        basic.append(vibrate)

        autolock = RadioSetting("autolock", "Auto Lock",
                                RadioSettingValueBoolean(_settings.autolock))
        basic.append(autolock)

        calltone = RadioSetting("calltone", "Call Tone",
                                RadioSettingValueInteger(
                                    1, 10, _settings.calltone))
        basic.append(calltone)

        roger = RadioSetting("roger", "Roger Tone",
                             RadioSettingValueList(
                                 LIST_ROGER, LIST_ROGER[_settings.roger]))
        basic.append(roger)

        squelch = RadioSetting("squelch", "Squelch Level",
                               RadioSettingValueInteger(
                                   0, 10, _settings.squelch))
        basic.append(squelch)

        def apply_tot_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(
                      setting.value) + " from list")
            val = str(setting.value)
            index = TOT_CHOICES.index(val)
            val = TOT_VALUES[index]
            obj.set_value(val)

        if _settings.tot in TOT_VALUES:
            idx = TOT_VALUES.index(_settings.tot)
        else:
            idx = TOT_VALUES.index(0x78)
        rs = RadioSettingValueList(TOT_CHOICES, TOT_CHOICES[idx])
        rset = RadioSetting("tot", "Time-out Timer", rs)
        rset.set_apply_callback(apply_tot_listvalue, _settings.tot)
        basic.append(rset)

        # Side Key Settings
        def apply_skey_listvalue(setting, obj):
            LOG.debug("Setting value: " + str(
                      setting.value) + " from list")
            val = str(setting.value)
            index = SKEY_CHOICES.index(val)
            val = SKEY_VALUES[index]
            obj.set_value(val)

        # Side Key (Upper) - Short Press
        if _settings.skeyus in SKEY_VALUES:
            idx = SKEY_VALUES.index(_settings.skeyus)
        else:
            idx = SKEY_VALUES.index(0x01)
        rs = RadioSettingValueList(SKEY_CHOICES, SKEY_CHOICES[idx])
        rset = RadioSetting("skeyus", "Side Key(upper) - Short Press", rs)
        rset.set_apply_callback(apply_skey_listvalue, _settings.skeyus)
        sidekey.append(rset)

        # Side Key (Upper) - Long Press
        if _settings.skeyul in SKEY_VALUES:
            idx = SKEY_VALUES.index(_settings.skeyul)
        else:
            idx = SKEY_VALUES.index(0x04)
        rs = RadioSettingValueList(SKEY_CHOICES, SKEY_CHOICES[idx])
        rset = RadioSetting("skeyul", "Side Key(upper) - Long Press", rs)
        rset.set_apply_callback(apply_skey_listvalue, _settings.skeyul)
        sidekey.append(rset)

        # Side Key (Lower) - Short Press
        if _settings.skeyds in SKEY_VALUES:
            idx = SKEY_VALUES.index(_settings.skeyds)
        else:
            idx = SKEY_VALUES.index(0x03)
        rs = RadioSettingValueList(SKEY_CHOICES, SKEY_CHOICES[idx])
        rset = RadioSetting("skeyds", "Side Key(lower) - Short Press", rs)
        rset.set_apply_callback(apply_skey_listvalue, _settings.skeyds)
        sidekey.append(rset)

        # Side Key (Lower) - Long Press
        if _settings.skeyul in SKEY_VALUES:
            idx = SKEY_VALUES.index(_settings.skeydl)
        else:
            idx = SKEY_VALUES.index(0x14)
        rs = RadioSettingValueList(SKEY_CHOICES, SKEY_CHOICES[idx])
        rset = RadioSetting("skeydl", "Side Key(lower) - Long Press", rs)
        rset.set_apply_callback(apply_skey_listvalue, _settings.skeydl)
        sidekey.append(rset)

        # VOX Settings
        vox = RadioSetting("vox", "VOX",
                           RadioSettingValueBoolean(_settings.vox))
        voxset.append(vox)

        voxl = RadioSetting("voxl", "VOX Level",
                            RadioSettingValueInteger(
                                0, 10, _settings.voxl))
        voxset.append(voxl)

        voxd = RadioSetting("voxd", "VOX Delay (seconde)",
                            RadioSettingValueList(
                                LIST_VOXD, LIST_VOXD[_settings.voxd]))
        voxset.append(voxd)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RB15Radio(RB15RadioBase):
    """RETEVIS RB15"""
    VENDOR = "Retevis"
    MODEL = "RB15"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _ranges = [
               (0x0150, 0x07A0),
              ]
    _memsize = 0x07A0

    _upper = 99
    _frs = False  # sold as FRS radio but supports full band TX/RX


@directory.register
class RB615RadioBase(RB15RadioBase):
    """RETEVIS RB615"""
    VENDOR = "Retevis"
    MODEL = "RB615"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _ranges = [
               (0x0150, 0x07A0),
              ]
    _memsize = 0x07A0

    _upper = 99
    _pmr = False  # sold as PMR radio but supports full band TX/RX



# Copyright 2013 Sean Burford <sburford@google.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import bandplan

SHORTNAME = "iaru_r3"

DESC = {
  "name": "IARU Region 3 (Asia Pacific)",
  "updated": "16 October 2009",
  "url": "http://www.iaru.org/uploads/1/3/0/7/13073366/r3_band_plan.pdf"
}

# Bands are broken up like this so that other plans can import bits.

BANDS_2100M = (
  bandplan.Band((135700, 137800), "137 kHz Band", mode="CW"),
)

BANDS_160M = (
  bandplan.Band((1800000, 2000000), "160 Meter Band"),
  bandplan.Band((1830000, 1840000), "Digimodes", mode="RTTY"),
  bandplan.Band((1840000, 2000000), "Phone"),
)

BANDS_80M = (
  bandplan.Band((3500000, 3900000), "80 Meter Band"),
  bandplan.Band((3500000, 3510000), "CW, priority for DX", mode="CW"),
  bandplan.Band((3535000, 3900000), "Phone"),
  bandplan.Band((3775000, 3800000), "All modes, SSB DX preferred", mode="LSB"),
)

BANDS_40M = (
  bandplan.Band((7000000, 7300000), "40 Meter Band"),
  bandplan.Band((7000000, 7025000), "CW, priority for DX", mode="CW"),
  bandplan.Band((7025000, 7035000), "All narrow band modes, cw", mode="CW"),
  bandplan.Band((7035000, 7040000), "All narrow band modes, phone"),
  bandplan.Band((7040000, 7300000), "All modes, digimodes"),
)

BANDS_30M = (
  bandplan.Band((10100000, 10150000), "30 Meter Band"),
  bandplan.Band((10100000, 10130000), "CW", mode="CW"),
  bandplan.Band((10130000, 10150000), "All narrow band digimodes"),
)

BANDS_20M = (
  bandplan.Band((14000000, 14350000), "20 Meter Band"),
  bandplan.Band((14000000, 14070000), "CW", mode="CW"),
  bandplan.Band((14070000, 14099000), "All narrow band modes, digimodes"),
  bandplan.Band((14099000, 14101000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((14101000, 14112000), "All narrow band modes, digimodes"),
  bandplan.Band((14101000, 14350000), "All modes, digimodes"),
)

BANDS_17M = (
  bandplan.Band((18068000, 18168000), "17 Meter Band"),
  bandplan.Band((18068000, 18100000), "CW", mode="CW"),
  bandplan.Band((18100000, 18109000), "All narrow band modes, digimodes"),
  bandplan.Band((18109000, 18111000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((18111000, 18168000), "All modes, digimodes"),
)

BANDS_15M = (
  bandplan.Band((21000000, 21450000), "15 Meter Band"),
  bandplan.Band((21000000, 21070000), "CW", mode="CW"),
  bandplan.Band((21070000, 21125000), "All narrow band modes, digimodes"),
  bandplan.Band((21125000, 21149000), "All narrow band modes, digimodes"),
  bandplan.Band((21149000, 21151000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((21151000, 21450000), "All modes", mode="USB"),
)

BANDS_12M = (
  bandplan.Band((24890000, 24990000), "12 Meter Band"),
  bandplan.Band((24890000, 24920000), "CW", mode="CW"),
  bandplan.Band((24920000, 24929000), "All narrow band modes, digimodes"),
  bandplan.Band((24929000, 24931000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((24931000, 24990000), "All modes, digimodes", mode="USB"),
)

BANDS_10M = (
  bandplan.Band((28000000, 29700000), "10 Meter Band"),
  bandplan.Band((28000000, 28050000), "CW", mode="CW"),
  bandplan.Band((28050000, 28150000), "All narrow band modes, digimodes"),
  bandplan.Band((28150000, 28190000), "All narrow band modes, digimodes"),
  bandplan.Band((28190000, 28199000), "Beacons", mode="CW"),
  bandplan.Band((28199000, 28201000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((28201000, 28300000), "Beacons", mode="CW"),
  bandplan.Band((28300000, 29300000), "Phone"),
  bandplan.Band((29300000, 29510000), "Satellite downlink"),
  bandplan.Band((29510000, 29520000), "Guard band, no transmission allowed"),
  bandplan.Band((29520000, 29700000), "Wide band", step_khz=10, mode="NFM"),
)

BANDS_6M = (
  bandplan.Band((50000000, 54000000), "6 Meter Band"),
  bandplan.Band((50000000, 50100000), "Beacons", mode="CW"),
  bandplan.Band((50100000, 50500000), "Phone and narrow band"),
  bandplan.Band((50500000, 54000000), "Wide band"),
)

BANDS_2M = (
  bandplan.Band((144000000, 148000000), "2 Meter Band"),
  bandplan.Band((144000000, 144035000), "Earth Moon Earth"),
  bandplan.Band((145800000, 146000000), "Satellite"),
)

BANDS_70CM = (
  bandplan.Band((430000000, 450000000), "70cm Band"),
  bandplan.Band((431900000, 432240000), "Earth Moon Earth"),
  bandplan.Band((435000000, 438000000), "Satellite"),
)

BANDS_23CM = (
  bandplan.Band((1240000000, 1300000000), "23cm Band"),
  bandplan.Band((1260000000, 1270000000), "Satellite"),
  bandplan.Band((1296000000, 1297000000), "Earth Moon Earth"),
)

BANDS = BANDS_2100M + BANDS_160M + BANDS_80M + BANDS_40M + BANDS_30M + \
        BANDS_20M + BANDS_17M + BANDS_15M + BANDS_12M + BANDS_10M + \
        BANDS_6M + BANDS_2M + BANDS_70CM + BANDS_23CM
